"""Helper functions for the text `rebuilder.py` script."""

import json
import logging
import os
from typing import Any, Optional, Union

from impresso_commons.utils.s3 import (
    IMPRESSO_STORAGEOPT,
    alternative_read_text,
    get_s3_resource,
    get_s3_versions,
)

logger = logging.getLogger(__name__)

IIIF_ENDPOINT_BASE_2_SUFFIX = {
    # suffix for SWA data
    "https://ub-sipi.ub.unibas.ch/impresso": "max/0/default.jpg",
    # suffix for BCUL data, sometimes 300 can be replaced by larger values
    "https://scriptorium.bcu-lausanne.ch/api": "300,/0/default.jpg",
}

WHITESPACE_RULES = {
    "fr": {
        "pct_no_ws_before": [".", ",", ")", "]", "}", "°", "...", ".-", "%"],
        "pct_no_ws_after": ["(", "[", "{"],
        "pct_no_ws_before_after": ["'", "-"],
        "pct_number": [".", ","],
    },
    "de": {
        "pct_no_ws_before": [
            ".",
            ",",
            ")",
            "]",
            "}",
            "°",
            "...",
            "?",
            "!",
            ":",
            ";",
            ".-",
            "%",
        ],
        "pct_no_ws_after": ["(", "[", "{"],
        "pct_no_ws_before_after": ["'", "-"],
        "pct_number": [".", ","],
    },
    "other": {
        "pct_no_ws_before": [
            ".",
            ",",
            ")",
            "]",
            "}",
            "°",
            "...",
            "?",
            "!",
            ":",
            ";",
            ".-",
            "%",
        ],
        "pct_no_ws_after": ["(", "[", "{"],
        "pct_no_ws_before_after": ["'", "-"],
        "pct_number": [".", ","],
    },
}


def read_issue(issue, bucket_name, s3_client=None):
    """Read the data from S3 for a given newspaper issue.

    NB: It injects the s3_version into the returned object.

    :param issue: input issue
    :type issue: `IssueDir`
    :param bucket_name: bucket's name
    :type bucket_name: str
    :param s3_client: open connection to S3 storage
    :type s3_client: `boto3.resources.factory.s3.ServiceResource`
    :return: a JSON representation of the issue object
    """
    if s3_client is None:
        s3_client = get_s3_resource()

    content_object = s3_client.Object(bucket_name, issue.path)
    file_content = content_object.get()["Body"].read().decode("utf-8")
    issue_json = json.loads(file_content)
    issue_json["s3_version"] = get_s3_versions(bucket_name, issue.path)[0][0]
    logger.info("Read JSON of %s", issue)
    return (issue, issue_json)


def read_page(page_key, bucket_name, s3_client):
    """Read the data from S3 for a given newspaper pages."""

    try:
        content_object = s3_client.Object(bucket_name, page_key)
        file_content = content_object.get()["Body"].read().decode("utf-8")
        page_json = json.loads(file_content)
        page_json["s3v"] = get_s3_versions(bucket_name, page_key)[0][0]
        logger.info("Read page %s from bucket %s", page_key, bucket_name)
        return page_json
    except Exception as e:
        logger.error("There was a problem reading %s: %s", page_key, e)
        return None


def read_issue_pages(issue, issue_json, bucket=None):
    """Read all pages of a given issue from S3 in parallel."""
    newspaper = issue.journal
    year = issue.date.year

    filename = (
        f"{bucket}/{newspaper}/pages/{newspaper}-{year}"
        f"/{issue_json['id']}-pages.jsonl.bz2"
    )

    pages = [
        json.loads(page)
        for page in alternative_read_text(filename, IMPRESSO_STORAGEOPT)
    ]

    print(filename)
    issue_json["pp"] = pages
    del pages
    return (issue, issue_json)


def rejoin_articles(issue, issue_json):
    print(f"Rejoining pages for issue {issue.path}")
    articles = []
    for article in issue_json["i"]:

        art_id = article["m"]["id"]
        article["m"]["s3v"] = issue_json["s3_version"]
        article["has_problem"] = False
        article["m"]["pp"] = sorted(list(set(article["m"]["pp"])))

        pages = []
        page_ids = [page["id"] for page in issue_json["pp"]]
        for page_no in article["m"]["pp"]:
            # given a page  number (from issue.json) and its canonical ID
            # find the position of that page in the array of pages (with text
            # regions)
            page_no_string = f"p{str(page_no).zfill(4)}"
            try:
                page_idx = [
                    n
                    for n, page in enumerate(issue_json["pp"])
                    if page_no_string in page["id"]
                ][0]
                pages.append(issue_json["pp"][page_idx])
            except IndexError:
                article["has_problem"] = True
                articles.append(article)
                logger.error(
                    "Page %s not found for item %s. Issue %s has pages %s",
                    page_no_string,
                    art_id,
                    issue_json["id"],
                    page_ids,
                )
                continue

        regions_by_page = []
        for page in pages:
            regions_by_page.append(
                [
                    region
                    for region in page["r"]
                    if "pOf" in region and region["pOf"] == art_id
                ]
            )
        article["pprr"] = regions_by_page
        try:
            convert_coords = [p["cc"] for p in pages]
            article["m"]["cc"] = sum(convert_coords) / len(convert_coords) == 1.0
        except Exception:
            # it just means there was no CC field in the pages
            article["m"]["cc"] = None

        articles.append(article)
    return articles


def pages_to_article(article, pages):
    """Return all text regions belonging to a given article."""
    try:
        art_id = article["m"]["id"]
        print("Extracting text regions for article %s", art_id)
        regions_by_page = []
        for page in pages:
            regions_by_page.append(
                [region for region in page["r"] if region["pOf"] == art_id]
            )
        convert_coords = [page["cc"] for page in pages]
        article["m"]["cc"] = sum(convert_coords) / len(convert_coords) == 1.0
        article["has_problem"] = False
        article["pprr"] = regions_by_page
        return article
    except Exception:
        article["has_problem"] = True
        return article


def text_apply_breaks(fulltext, breaks):
    """Apply breaks to the text returned by `rebuild_for_solr`.

    The purpose of this function is to debug (visually) the `rebuild_for_solr`
    function. It applies to `fulltext` the characte offsets contained in
    `breaks` (e.g. line breaks, paragraph breaks, etc.).

    :param fulltext: input text
    :type fulltext: str
    :param breaks: a list of character offsets
    :type breaks: list of int
    :return: a list of text chunks
    :rtype: list
    """

    text = []
    start = 0

    for br in breaks:
        text.append(fulltext[start:br].strip())
        start = br

    text.append(fulltext[start:])

    return text


def get_iiif_and_coords(ci: dict[str, Any]) -> tuple[Optional[str], Optional[str]]:
    """Fetch the iiif link and image coordinates from CI metadata.

    Adapts to the various cases currently present in the canonical data, see
    https://github.com/impresso/impresso-text-acquisition/issues/117.

    Args:
        ci (dict[str, Any]): Content item to retrieve the information from.

    Returns:
        tuple[str | None, str | None]: IIIF link and coordinates as string or
            None if part of the information is missing from the content item
    """
    if "iiif_link" in ci or "iiif_link" in ci["m"]:
        iiif_link = ci["m"]["iiif_link"] if "iiif_link" in ci["m"] else ci["iiif_link"]

        if "c" in ci or "c" in ci["m"]:
            coords = ci["c"] if "c" in ci else ci["m"]["c"]

            if iiif_link and coords:
                return iiif_link, ",".join([str(c) for c in coords])
    return None, None


def reconstruct_iiif_link(content_item: dict[str, Any]) -> str:
    """Construct the iiif link to the CI's image based on its metadata.

    A iiif image API link and the image coordinates are to be fetched
    from the content item first.
    Different importers (and endpoints) have different formats, needing
    different processing.
    In addition, some inconsistencies exist in the canonical data.
    This function adapts to these variations, more details in issue:
    https://github.com/impresso/impresso-text-acquisition/issues/117

    Args:
        content_item (dict[str, Any]): Content item in canonical format.

    Returns:
        str: iiif link to the image area of the content item if present in the
            CI metadata, else None.
    """
    img_suffix = "full/0/default.jpg"

    iiif, coords = get_iiif_and_coords(content_item)
    if iiif:
        # recover the url base to which the image suffix should be appended
        uri_base, old_suffix = os.path.split(iiif)

        # SWA and BCUL data have a different image suffix than other endpoints
        for iiif_base, iiif_suffix in IIIF_ENDPOINT_BASE_2_SUFFIX.items():
            img_suffix = iiif_suffix if iiif_base in uri_base else img_suffix

        if old_suffix == "default.jpg":
            # iiif was already constructed according to needs.
            if coords in iiif and img_suffix in iiif:
                return iiif
            # uri was already of image, but not correct.
            uri_base = "/".join(uri_base.split("/")[:-3])
        elif old_suffix != "info.json":
            warn_msg = (
                f"Unexpected iiif url suffix: {old_suffix} "
                f"for CI with id: {content_item['id']}."
            )
            logger.warning(warn_msg)

        # reconstruct the final image link
        return os.path.join(uri_base, coords, img_suffix)
    return None


def insert_whitespace(
    token: str,
    next_t: Optional[str],
    prev_t: Optional[str],
    lang: Optional[str],
) -> bool:
    """Determine whether a whitespace should be inserted after a token.

    Args:
        token (str): Current token.
        next_t (str): Following token.
        prev_t (str): Previous token.
        lang (str): Language of text.

    Returns:
        bool: Whether a whitespace should be inserted after the `token`.
    """
    # if current token text is None, previous token's whitespace rule applies
    if token is None or len(token) == 0:
        return False

    wsrules = WHITESPACE_RULES[lang if lang in WHITESPACE_RULES else "other"]

    insert_ws = True

    if (
        token in wsrules["pct_no_ws_before_after"]
        or next_t in wsrules["pct_no_ws_before_after"]
    ):
        insert_ws = False

    # the first char of the next token is punctuation.
    elif next_t is not None and len(next_t) != 0:
        if (
            next_t in wsrules["pct_no_ws_before"]
            or next_t[0] in wsrules["pct_no_ws_before"]
        ):
            insert_ws = False

    # the last char of current token is punctuation.
    elif token in wsrules["pct_no_ws_after"] or token[-1] in wsrules["pct_no_ws_after"]:
        insert_ws = False

    elif token in wsrules["pct_number"] and prev_t is not None and next_t is not None:
        if prev_t.isdigit() and next_t.isdigit():
            return False
        else:
            return True

    debug_msg = (
        f"Insert whitespace: curr={token}, follow={next_t}, "
        f"prev={prev_t} ({insert_ws})"
    )
    logger.debug(debug_msg)

    return insert_ws



#!/usr/bin/python3
"""
Functions to support re-computation of Olive box coordinates.
"""

import logging
import cv2 as cv
from bs4 import BeautifulSoup
import os
import zipfile
import numpy as np

from impresso_commons.images import img_utils

__author__ = "maudehrmann"


logger = logging.getLogger(__name__)


def get_iiif_url(
    page_id: str,
    box: str,
    base: str = "http://dhlabsrv17.epfl.ch/iiif_impresso",
    iiif_manifest_uri: str = None,
    pct: bool = False,
) -> str:
    """ Returns impresso iiif url given a page id and a box

    :param page_id: impresso page id, e.g. EXP-1930-06-10-a-p0001
    :type page_id: str
    :param box: iiif box (x, y, w, h)
    :type box: str (4 coordinate values blank separated)
    :return: iiif url of the box
    :rtype: str
    """
    prefix = "pct:" if pct else ""
    suffix = "full/0/default.jpg"

    if iiif_manifest_uri is None:
        return os.path.join(base, page_id, prefix + box.replace(" ", ","), suffix)
    else:
        return os.path.join(iiif_manifest_uri.replace('/info.json', ''), prefix + box.replace(" ", ","), suffix)


def compute_scale_factor(img_source_path, img_dest_path):
    """
    Computes x scale factor bewteen 2 images.

    :param img_source_path: the source image
    :type img_source_path: full path to the image
    :param img_dest_path: the destination image
    :type img_dest_path: full path to the image
    :return: scale factor
    :rtype: float
    """
    img_s = cv.imread(img_source_path, 0)
    img_d = cv.imread(img_dest_path, 0)
    x_s = img_s.shape[0]
    x_d = img_d.shape[0]
    return x_d / x_s


def compute_box(scale_factor, input_box):
    """
    Compute IIIF box coordinates of input_box relative to scale_factor.

    :param scale_factor: ratio between 2 images with different dimensions
    :type scale_factor: float
    :param input_box: string with 4 values separated by spaces
    :type input_box: str
    :return: new box coordinates
    :rtype: str
    """
    try:
        elems = input_box.split(" ")
    except ValueError:
        logger.info(f'Invalid box format: {input_box}')
        return

    x = round(int(elems[0]) * scale_factor)
    y = round(int(elems[1]) * scale_factor)
    w = round((int(elems[2]) - int(elems[0])) * scale_factor)
    h = round((int(elems[3]) - int(elems[1])) * scale_factor)
    return " ".join([str(x), str(y), str(w), str(h)])


def convert_box(input_box):
    """
    Convert a box with [x y x y] coordinates to [x y w h]

    :param input_box: box with 4 coordinates, x and y upper left and lower right
    :type input_box: str
    :return: box with 4 coordinates, x and y upper left, width and height
    :rtype: str
    """
    try:
        elems = input_box.split(" ")
    except ValueError:
        logger.info(f'Invalid box format: {input_box}')
        return

    w = int(elems[2]) - int(elems[0])
    h = int(elems[3]) - int(elems[1])
    return ",".join([elems[0], elems[1], str(w), str(h)])


def get_scale_factor(issue_dir_path, archive, page_xml, box_strategy, img_source_name):
    """
    Returns the scale factor in Olive context, given a strategy to choose the source image.

    :param issue_dir_path: the path of the issue
    :type  issue_dir_path: str
    :param archive: the zip archive
    :type  archive: zipfile.ZipFile
    :param page_xml: the xml handler of the page
    :type page_xml: bytes
    :param box_strategy: the box strategy such as found in the info.txt from jp2 folder
    :type box_strategy: str
    :param img_source_name: as found in the info.txt from jp2 folder
    :return: the hopefully correct scale factor
    :rtype: float

    Background information
    ======================
    Impresso converts library images to JP2, taking the best image available: tif > highest png > jpg.
    Olive box coordinates were computed according to an image source which we have to identify among several.
    Image format coverage is different from issue to issue, and we have to devise strategies.

    Case 1: tif
    -----------
    The tif is present and is the file from which the jp2 was converted.
    Dest: Tif dimensions can therefore be used as jp2 dimensions, no need to read the jp2 file.
    Source: Image source dimension is present in the page.xml (normally).

    Case 2: several png
    -------------------
    In this case the jp2 was acquired using the png with the highest dimension.
    Dest: It looks that in case of several png, Olive also took the highest for the OCR. It is therefore
    possible to rely on the resolution indicated in the page xml, which should be the same as our jp2.
    N.B.: the page width and heigth indicated in the xml do not correspond (usually) to the highest
    resolution png (there is therefore a discrepancy in Olive file between the tag 'images_resolution'
    on the one hand, and 'page_width|height'on the other). It seems we can ignore this and rely on the
    resolution only in the current case.
    Source: the highest png
    Here source and dest dimension are equals, the function returns 1.

    Case 3: one png only
    --------------------
    To be checked if it happens.
    In this case, there is no choice and Olive OCR and JP2 acquisition should be from the same source
    => scale factor of 1.
    Here we do an additional check to see if the page_width|height are the same as the image ones.
    The only danger is if Olive used another image file and did not provide it.

    Case 4: one jpg only
    --------------------
    Same as Case 3, scale factor of 1.
    Here we do an additional check to see if the page_width|height are the same as the image ones.
    (there is only one image and things should fit, not like in case 2)
    """
    page_soup = BeautifulSoup(page_xml, "lxml")
    page_root = page_soup.find("xmd-page")
    page_number = page_root.meta["page_no"]
    if box_strategy == img_utils.BoxStrategy.tif.name:
        for f in page_root.datafiles.find_all('files'):
            if f['type'] == 'PAGE_IMG' and f['present'] == "1":
                source_res = f['xresolution_dpi']
                dest_res = page_root.meta['images_resolution']
                break
        if source_res and dest_res:
            return int(source_res) / int(dest_res)

        else:
            logger.info(f"Impossible to get resolution in case: tif" " in {issue_dir_path}, page {page_number}")
            return None

    elif box_strategy == img_utils.BoxStrategy.png_highest.name:
        if "_" not in img_source_name:
            logger.info(f"Not valid png filename {img_source_name}")
            return None

        png_res = os.path.splitext(img_source_name)[0].split("_", 1)[-1]
        olive_res = page_root.meta['images_resolution']
        if png_res == olive_res:
            return 1.0
        else:
            logger.info(
                f"Incompatible resolutions between highest png and olive indications \
            in {issue_dir_path}, page {page_number}"
            )
            print(
                f"Incompatible resolutions between highest png and olive indications \
                        in {issue_dir_path}, page {page_number}"
            )
            return None

    elif box_strategy == img_utils.BoxStrategy.png_uniq.name:
        # TODO if needed
        logger.info(f"Finally found a case of {img_utils.BoxStrategy.png_uniq}, "
                    "which is not ready yet")

    elif box_strategy == img_utils.BoxStrategy.jpg_uniq.name:
        # get the x dimension of the unique jpg (from which jp2 was acquired)
        # and compare with olive's one
        img_data = archive.read(img_source_name)
        img = cv.imdecode(np.frombuffer(img_data, np.uint8), 1)
        jpg_x_dim = img.shape[1]
        olive_x_dim = page_root.meta['page_width']
        if jpg_x_dim == int(olive_x_dim):
            return 1.0
        else:
            logger.info(
                "Incompatible resolutions between uniq jpg and olive indications"
                " in {issue_dir_path}, page {page_number}."
            )
            return None


def test():
    """ a testing function - to be moved in proper test unit..."""
    base_dir = "/Users/maudehrmann/work/work-projects/impresso/code/impresso-image-acquisition/sample-data"

    # tifs
    # box = "262 624 335 650"  # 'Centenaire' in p1 Ar00100.xml
    # archive = os.path.join(base_dir, "TEST/1900/01/10/Document.zip")
    # working_archive = zipfile.ZipFile(archive)
    # data = working_archive.read("1/Pg001.xml")
    # sf = get_scale_factor("fictious path", working_archive, data, img_utils.BoxStrategy.tif.name, None)
    # newbox = compute_box(sf, box)
    # iiif = get_iiif_url("GDL-1900-01-10-a-p0001", newbox)
    # print("\nCASE: tif - word 'Centenaire'")
    # print(f"Scale factor: {sf}")
    # print(f"Newbox: {newbox}")
    # print(f"IIIF: {iiif}")

    # jpg_uniq
    # box = "483 502 556 517"  # Hambourg in p1 Ar00100.xml
    # archive = os.path.join(base_dir, "TEST/1900/01/12/Document.zip")
    # working_archive = zipfile.ZipFile(archive)
    # page_data = working_archive.read("1/Pg001.xml")
    # sf = get_scale_factor("fictious path", working_archive, page_data, img_utils.BoxStrategy.jpg_uniq.name, "1/Img/Pg001.jpg")
    # newbox = compute_box(sf, box)
    # iiif = get_iiif_url("LCE-1868-08-02-a-p0001", newbox)
    # print("\nCASE: jpj uniq - word 'hambourg'")
    # print(f"Scale factor: {sf}")
    # print(f"Newbox: {newbox}")
    # print(f"IIIF: {iiif}")

    # JDG 1973-10-06 tif
    # box = "282 85 320 1007"  # NEW in p6 Ar00600.xml
    # archive = os.path.join(base_dir, "TEST/JDG-1973-10-06/Document.zip")
    # working_archive = zipfile.ZipFile(archive)
    # page_data = working_archive.read("6/Pg006.xml")
    # sf = get_scale_factor("fictious path", working_archive, page_data, img_utils.BoxStrategy.tif.name, None)
    # newbox = compute_box(sf, box)
    # iiif = get_iiif_url("JDG-1973-10-06-a-p0006", newbox)
    # print("\nCASE: tif - word 'NEW'")
    # print(f"Scale factor: {sf}")
    # print(f"Oldbox: {box}")
    # print(f"Newbox: {newbox}")
    # print(f"IIIF: {iiif}")

    # # JDG 1973-10-01 tif
    # box = "132 467 268 504"  # Bonnes in p1 Ar00100.xml
    # archive = os.path.join(base_dir, "TEST/JDG-1973-10-06/Document.zip")
    # working_archive = zipfile.ZipFile(archive)
    # page_data = working_archive.read("1/Pg001.xml")
    # sf = get_scale_factor("fictious path", working_archive, page_data, img_utils.BoxStrategy.tif.name, None)
    # newbox = compute_box(sf, box)
    # iiif = get_iiif_url("JDG-1973-10-06-a-p0001", newbox)
    # print("\nCASE: tif - word 'Bonnes'")
    # print(f"Scale factor: {sf}")
    # print(f"Oldbox: {box}")
    # print(f"Newbox: {newbox}")
    # print(f"IIIF: {iiif}")
    #
    # # png_highest
    # box = "1047 1006 1173 1036"  # 'NEUCHATEL' in p1 Ad00103.xml
    # archive = os.path.join(base_dir, "TEST/1900/01/11/Document.zip")
    # working_archive = zipfile.ZipFile(archive)
    # data = working_archive.read("1/Pg001.xml")
    # sf = get_scale_factor("fictious path", working_archive, data, img_utils.BoxStrategy.png_highest.name,
    #                       "Img/Pg001_180.png")
    # print(sf)
    # newbox = compute_box(sf, box)
    # iiif = get_iiif_url("EXP-1889-07-01-a-p0001", newbox)
    # print(f"CASE: png_highest - 'NEUCHATEL'\nScale factor: {sf}\nNewbox: {newbox}\nIIIF: {iiif}\n")
    #
    # # IMP-1985-12-17;  pages 25,26,27 png highest
    # # box = "97 312 253 408"  # Cerf in p25 Ar02500.xml
    # # archive = os.path.join(base_dir, "TEST/IMP-1985-12-17/Document.zip")
    # # working_archive = zipfile.ZipFile(archive)
    # # page_data = working_archive.read("25/Pg025.xml")
    # # sf = get_scale_factor("fictious path", working_archive, page_data, img_utils.BoxStrategy.png_highest.name,
    # #                       "Img/Pg025_100.png")
    # # newbox = compute_box(sf, box)
    # # iiif = get_iiif_url("IMP-1985-12-17-a-p0025", newbox)
    # # print("CASE: png highest - word 'Cerf'")
    # # print(f"Scale factor: {sf}")
    # # print(f"Oldbox: {box}")
    # # print(f"Newbox: {newbox}")
    # # print(f"IIIF: {iiif}\n")
    #
    # # JDG 1877-12-08;  pages 1,2,3,4
    # box = "212 555 315 576"  # graphique in p1 Ar00104.xml
    # archive = os.path.join(base_dir, "TEST/JDG-1877-12-08/Document.zip")
    # working_archive = zipfile.ZipFile(archive)
    # page_data = working_archive.read("1/Pg001.xml")
    # sf = get_scale_factor("fictious path", working_archive, page_data, img_utils.BoxStrategy.png_highest.name,
    #                       "Img/Pg001_120.png")
    # newbox = compute_box(sf, box)
    # iiif = get_iiif_url("JDG-1877-12-08-a-p0001", newbox)
    # print("CASE: png highest - word 'graphique'")
    # print(f"Scale factor: {sf}")
    # print(f"Oldbox: {box}")
    # print(f"Newbox: {newbox}")
    # print(f"IIIF: {iiif}\n")

    # JDG 1887-08-26;  pages 1,2,3,4
    # box = "1568 579 1612 595"  # terrain in p1 Ar00108.xml
    # archive = os.path.join(base_dir, "TEST/JDG-1887-08-26/Document.zip")
    # working_archive = zipfile.ZipFile(archive)
    # page_data = working_archive.read("1/Pg001.xml")
    # sf = get_scale_factor("fictious path", working_archive, page_data, img_utils.BoxStrategy.png_highest.name,
    #                       "Img/Pg001_60.png")
    # newbox = compute_box(sf, box)
    # iiif = get_iiif_url("JDG-1887-08-26-a-p0001", newbox)
    # print("CASE: png highest - word 'terrain'")
    # print(f"Scale factor: {sf}")
    # print(f"Oldbox: {box}")
    # print(f"Newbox: {newbox}")
    # print(f"IIIF: {iiif}\n")

    # JDG 1860-02-08;  pages 1,2,3,4 png highest
    # box = "580 77 625 99"  # POST in p1 Ar00104.xml
    # archive = os.path.join(base_dir, "TEST/JDG-1860-02-08/Document.zip")
    # working_archive = zipfile.ZipFile(archive)
    # page_data = working_archive.read("1/Pg001.xml")
    # sf = get_scale_factor("fictious path", working_archive, page_data, img_utils.BoxStrategy.png_highest.name,
    #                       "Img/Pg001_120.png")
    # newbox = compute_box(sf, box)
    # iiif = get_iiif_url("JDG-1860-02-08-a-p0001", newbox)
    # print("CASE: png highest - word 'POST'")
    # print(f"Scale factor: {sf}")
    # print(f"Oldbox: {box}")
    # print(f"Newbox: {newbox}")
    # print(f"IIIF: {iiif}\n")

    # LBP/1880/04/30;  pages 1,2,3,4 png highest
    # box = "1616 416 1859 513"  # Dimanche in p1 Ar00104.xml
    # archive = os.path.join(base_dir, "TEST/LBP-1880-04-30/Document.zip")
    # working_archive = zipfile.ZipFile(archive)
    # page_data = working_archive.read("1/Pg001.xml")
    # sf = get_scale_factor("fictious path", working_archive, page_data, img_utils.BoxStrategy.png_highest.name,
    #                       "Img/Pg001_60.png")
    # newbox = compute_box(sf, box)
    # iiif = get_iiif_url("LBP-1880-04-30-a-p0001", newbox)
    # print("CASE: png highest - word 'Dimanche'")
    # print(f"Scale factor: {sf}")
    # print(f"Oldbox: {box}")
    # print(f"Newbox: {newbox}")
    # print(f"IIIF: {iiif}\n")

    # JDV-1848-05-24;  pages 1,2,3,4 jpg uniq
    box = "62 428 99 443"  # PRIX in p1 Ar00100.xml
    archive = os.path.join(base_dir, "TEST/JDV-1848-05-24/Document.zip")
    working_archive = zipfile.ZipFile(archive)
    page_data = working_archive.read("1/Pg001.xml")
    sf = get_scale_factor(
        "fictious path", working_archive, page_data, img_utils.BoxStrategy.jpg_uniq.name, "1/Img/Pg001.jpg"
    )

    newbox = compute_box(sf, box)
    iiif = get_iiif_url("JDV-1848-05-24-a-p0001", newbox)
    print("CASE: jpg uniq - word 'PRIX'")
    print(f"Scale factor: {sf}")
    print(f"Oldbox: {box}")
    print(f"Newbox: {newbox}")
    print(f"IIIF: {iiif}\n")

    # JDV-1848-05-24;  pages 1,2,3,4 jpg uniq
    box = "557 1461 597 1480"  # intro in p1 Ar00101.xml
    archive = os.path.join(base_dir, "TEST/JDV-1848-05-24/Document.zip")
    working_archive = zipfile.ZipFile(archive)
    page_data = working_archive.read("1/Pg001.xml")
    sf = get_scale_factor(
        "fictious path", working_archive, page_data, img_utils.BoxStrategy.jpg_uniq.name, "1/Img/Pg001.jpg"
    )

    newbox = compute_box(sf, box)
    iiif = get_iiif_url("JDV-1848-05-24-a-p0001", newbox)
    print("CASE: jpg uniq - word 'intro'")
    print(f"Scale factor: {sf}")
    print(f"Oldbox: {box}")
    print(f"Newbox: {newbox}")
    print(f"IIIF: {iiif}\n")

    # JDV-1848-05-24;  pages 1,2,3,4 jpg uniq
    box = "517 1461 539 1480"  # être in p1 Ar00101.xml
    archive = os.path.join(base_dir, "TEST/JDV-1848-05-24/Document.zip")
    working_archive = zipfile.ZipFile(archive)
    page_data = working_archive.read("1/Pg001.xml")
    sf = get_scale_factor(
        "fictious path", working_archive, page_data, img_utils.BoxStrategy.jpg_uniq.name, "1/Img/Pg001.jpg"
    )

    newbox = compute_box(sf, box)
    iiif = get_iiif_url("JDV-1848-05-24-a-p0001", newbox)
    print("CASE: jpg uniq - word 'être'")
    print(f"Scale factor: {sf}")
    print(f"Oldbox: {box}")
    print(f"Newbox: {newbox}")
    print(f"IIIF: {iiif}\n")

    # JDV-1848-05-24;  pages 1,2,3,4 jpg uniq
    box = "751 1606 817 1625"  # collèges in p1 Ar00102.xml
    archive = os.path.join(base_dir, "TEST/JDV-1848-05-24/Document.zip")
    working_archive = zipfile.ZipFile(archive)
    page_data = working_archive.read("1/Pg001.xml")
    sf = get_scale_factor(
        "fictious path", working_archive, page_data, img_utils.BoxStrategy.jpg_uniq.name, "1/Img/Pg001.jpg"
    )

    newbox = compute_box(sf, box)
    iiif = get_iiif_url("JDV-1848-05-24-a-p0001", newbox)
    print("CASE: jpg uniq - word 'collèges'")
    print(f"Scale factor: {sf}")
    print(f"Oldbox: {box}")
    print(f"Newbox: {newbox}")
    print(f"IIIF: {iiif}\n")

    # EXP/1964/05/19/;  pages 5 png highest
    box = "142 254 228 277"  # dimensions in p5 Ar00500.xml
    archive = os.path.join(base_dir, "TEST/EXP-1964-05-19/Document.zip")
    working_archive = zipfile.ZipFile(archive)
    page_data = working_archive.read("5/Pg005.xml")
    sf = get_scale_factor(
        "fictious path", working_archive, page_data, img_utils.BoxStrategy.png_highest.name, "Img/Pg005_180.png"
    )
    print(sf)
    newbox = compute_box(sf, box)
    iiif = get_iiif_url("EXP-1964-05-19-a-p0005", newbox)
    print("CASE: png highest - word 'dimensions'")
    print(f"Scale factor: {sf}")
    print(f"Oldbox: {box}")
    print(f"Newbox: {newbox}")
    print(f"IIIF: {iiif}\n")

    # EXP/1964/05/19/;  pages 5 png highest
    box = "459 598 769 716"  # dimensions in p5 Ar00500.xml
    archive = os.path.join(base_dir, "TEST/EXP-1964-05-19/Document.zip")
    working_archive = zipfile.ZipFile(archive)
    page_data = working_archive.read("1/Pg001.xml")
    sf = get_scale_factor(
        "fictious path", working_archive, page_data, img_utils.BoxStrategy.png_highest.name, "Img/Pg001_180.png"
    )
    print(sf)
    newbox = compute_box(sf, box)
    iiif = get_iiif_url("EXP-1964-05-19-a-p0001", newbox)
    print("CASE: png highest - word 'dimensions'")
    print(f"Scale factor: {sf}")
    print(f"Oldbox: {box}")
    print(f"Newbox: {newbox}")
    print(f"IIIF: {iiif}\n")


if __name__ == '__main__':
    test()
    # box = "1096 454 1200 470"
    # print(convert_box(box))



"""Code for parsing impresso's canonical directory structures."""

import os
import logging
from datetime import date, datetime
from collections import namedtuple
import re

logger = logging.getLogger(__name__)

# a simple data structure to represent input directories
# a `Document.zip` file is expected to be found in `IssueDir.path`
IssueDir = namedtuple("IssueDir", ["journal", "date", "edition", "path"])

ContentItem = namedtuple(
    "Item", ["journal", "date", "edition", "number", "path", "type"]
)


KNOWN_JOURNALS = [
    "BDC",
    "CDV",
    "DLE",
    "EDA",
    "EXP",
    "IMP",
    "GDL",
    "JDF",
    "JDV",
    "LBP",
    "LCE",
    "LCG",
    "LCR",
    "LCS",
    "LES",
    "LNF",
    "LSE",
    "LSR",
    "LTF",
    "LVE",
    "EVT",
    "JDG",
    "LNQ",
    "NZZ",
    "FedGazDe",
    "FedGazFr",
    "FedGazIt",
    "arbeitgeber",
    "handelsztg",
    "actionfem",
    "armeteufel",
    "avenirgdl",
    "buergerbeamten",
    "courriergdl",
    "deletz1893",
    "demitock",
    "diekwochen",
    "dunioun",
    "gazgrdlux",
    "indeplux",
    "kommmit",
    "landwortbild",
    "lunion",
    "luxembourg1935",
    "luxland",
    "luxwort",
    "luxzeit1844",
    "luxzeit1858",
    "obermosel",
    "onsjongen",
    "schmiede",
    "tageblatt",
    "volkfreu1869",
    "waechtersauer",
    "waeschfra",
    "BLB",
    "BNN",
    "DFS",
    "DVF",
    "EZR",
    "FZG",
    "HRV",
    "LAB",
    "LLE",
    "MGS",
    "NTS",
    "NZG",
    "SGZ",
    "SRT",
    "WHD",
    "ZBT",
    "CON",
    "DTT",
    "FCT",
    "GAV",
    "GAZ",
    "LLS",
    "OIZ",
    "SAX",
    "SDT",
    "SMZ",
    "VDR",
    "VHT",
]


def pair_issue(issue_list1, issue_list2):
    """Associates pairs of issues originating from original and canonical repositories.

    :param issue_list1: list of IssueDir
    :type issue_list1: array
    :param issue_list2: list of IssueDir
    :type issue_list2: array
    :return: list containing tuples of issue pairs [(issue1, issue2), (...)]
    :rtype: list
    """
    dict1 = {}
    pairs = []
    for i in issue_list1:
        s_i = "-".join([i[0], str(i[1]), i[2]])
        dict1[s_i] = i

    for j in issue_list2:
        s_j = "-".join([j[0], str(j[1]), j[2]])
        if s_j in dict1:
            pairs.append((dict1[s_j], j))
    return pairs


def canonical_path(dir, name=None, extension=None, path_type="file"):
    """Create a canonical dir/file path from an `IssueDir` object.

    :param dir: an object representing a newspaper issue
    :type dir: `IssueDir`
    :param name: the file name (used only if path_type=='file')
    :type name: string
    :param extension: the file extension (used only if path_type=='file')
    :type extension: string
    :param path_type: type of path to build ('dir' | 'file')
    :type path_type: string
    :rtype: string
    """
    sep = "-" if path_type == "file" else "/"
    extension = extension if extension is not None else ""
    if path_type == "file":
        return "{}{}".format(
            sep.join(
                [
                    dir.journal,
                    str(dir.date.year),
                    str(dir.date.month).zfill(2),
                    str(dir.date.day).zfill(2),
                    dir.edition,
                    name,
                ]
            ),
            extension,
        )
    else:
        return sep.join(
            [
                dir.journal,
                str(dir.date.year),
                str(dir.date.month).zfill(2),
                str(dir.date.day).zfill(2),
                dir.edition,
            ]
        )


def _apply_datefilter(filter_dict, issues, year_only):
    filtered_issues = []

    for newspaper, dates in filter_dict.items():
        # date filter is a range
        if isinstance(dates, str):
            start, end = dates.split("-")
            start = datetime.strptime(start, "%Y/%m/%d").date()
            end = datetime.strptime(end, "%Y/%m/%d").date()

            if year_only:
                filtered_issues += [
                    i
                    for i in issues
                    if i.journal == newspaper and start.year <= i.date.year <= end.year
                ]
            else:
                filtered_issues += [
                    i
                    for i in issues
                    if i.journal == newspaper and start <= i.date <= end
                ]

        # date filter is not a range
        elif isinstance(dates, list):
            if not dates:
                filtered_issues += [i for i in issues if i.journal == newspaper]
            else:
                filter_date = [
                    (
                        datetime.strptime(d, "%Y/%m/%d").date().year
                        if year_only
                        else datetime.strptime(d, "%Y/%m/%d").date()
                    )
                    for d in dates
                ]

                if year_only:
                    filtered_issues += [
                        i
                        for i in issues
                        if i.journal == newspaper and i.date.year in filter_date
                    ]
                else:
                    filtered_issues += [
                        i
                        for i in issues
                        if i.journal == newspaper and i.date in filter_date
                    ]

    return filtered_issues


def select_issues(config_dict, inp_dir):
    """Reads a configuration file and select newspapers/issues to consider
    See config.example.md for explanations.

    Usage example:
        if config_file and os.path.isfile(config_file):
            with open(config_file, 'r') as f:
                config = json.load(f)
                issues = select_issues(config, inp_dir)
            else:
                issues = detect_issues(inp_dir)

    :param config_dict: dict of newspaper filter parameters
    :type config_dict: dict
    :param inp_dir: base dit where to get the issues from
    :type inp_dir: str

    """
    # read filters from json configuration (see config.example.json)
    try:
        filter_dict = config_dict.get("newspapers")
        exclude_list = config_dict["exclude_newspapers"]
        year_flag = config_dict["year_only"]
    except KeyError:
        logger.critical(
            f"The key [newspapers|exclude_newspapers|year_only] is missing in the config file."
        )
        return
    exclude_flag = False if not exclude_list else True
    logger.debug(
        f"got filter_dict: {filter_dict}, "
        f"\nexclude_list: {exclude_list}, "
        f"\nyear_flag: {year_flag}"
        f"\nexclude_flag: {exclude_flag}"
    )

    # detect issues to be imported
    if (
        not filter_dict and not exclude_list
    ):  # todo: remove this case? should be detect issue
        logger.debug(
            "No positive nor negative filter definition, all issues in {inp_dir} will be considered."
        )
        issues = detect_issues(inp_dir)
        return issues
    else:
        filter_newspapers = (
            set(filter_dict.keys()) if not exclude_list else set(exclude_list)
        )
        logger.debug(
            f"got filter_newspapers: {filter_newspapers}, with exclude flag: {exclude_flag}"
        )
        issues = detect_issues(
            inp_dir, journal_filter=filter_newspapers, exclude=exclude_flag
        )

        # apply date filter if not exclusion mode
        filtered_issues = (
            _apply_datefilter(filter_dict, issues, year_only=year_flag)
            if not exclude_flag
            else issues
        )
        return filtered_issues


def detect_issues(base_dir, journal_filter=None, exclude=False):
    """Parse a directory structure and detect newspaper issues to be imported.

    NB: invalid directories are skipped, and a warning message is logged.

    :param base_dir: the root of the directory structure
    :type base_dir: basestring
    :param journal_filter: list of newspaper to filter (positive or negative)
    :type journal_filter: set
    :param exclude: whether journal_filter is positive or negative
    :type exclude: boolean
    :rtype: list of `IssueDir` instances
    """
    detected_issues = []
    dir_path, dirs, files = next(os.walk(base_dir))
    # workaround to deal with journal-level folders like: 01_GDL, 02_GDL
    if journal_filter is None:
        journal_dirs = [d for d in dirs if d.split("_")[-1] in KNOWN_JOURNALS]
    else:
        if not exclude:
            filtrd_journals = list(set(KNOWN_JOURNALS).intersection(journal_filter))
        else:
            filtrd_journals = list(set(KNOWN_JOURNALS).difference(journal_filter))
        journal_dirs = [d for d in dirs if d.split("_")[-1] in filtrd_journals]

    for journal in journal_dirs:
        journal_path = os.path.join(base_dir, journal)
        journal = journal.split("_")[-1] if "_" in journal else journal
        dir_path, year_dirs, files = next(os.walk(journal_path))
        # year_dirs = [d for d in dirs if len(d) == 4]

        for year in year_dirs:
            year_path = os.path.join(journal_path, year)
            dir_path, month_dirs, files = next(os.walk(year_path))

            for month in month_dirs:
                month_path = os.path.join(year_path, month)
                dir_path, day_dirs, files = next(os.walk(month_path))

                for day in day_dirs:
                    day_path = os.path.join(month_path, day)
                    # concerning `edition="a"`: for now, no cases of newspapers
                    # published more than once a day in Olive format (but it
                    # may come later on)
                    try:
                        detected_issue = IssueDir(
                            journal,
                            date(int(year), int(month), int(day)),
                            "a",
                            day_path,
                        )
                        logger.debug("Found an issue: {}".format(str(detected_issue)))
                        detected_issues.append(detected_issue)
                    except ValueError:
                        logger.warning(
                            "Path {} is not a valid issue directory".format(day_path)
                        )
    return detected_issues


def detect_canonical_issues(base_dir, newspapers):
    """Parse a directory structure and detect newspaper issues to be imported.

    NB: invalid directories are skipped, and a warning message is logged.

    :param base_dir: the root of the directory structure
    :type base_dir: IssueDir
    :param newspapers: the list of newspapers to consider (acronym blank separated)
    :type newspapers: str
    :return: list of `IssueDir` instances
    :rtype: list
    """
    detected_issues = []
    dir_path, dirs, files = next(os.walk(base_dir))

    # workaround to deal with journal-level folders like: 01_GDL, 02_GDL
    # journal_dirs = [d for d in dirs if d.split("_")[-1] == newspapers]
    journal_dirs = [d for d in dirs if d in [d.split("_")[-1] for d in newspapers]]

    for journal in journal_dirs:
        journal_path = os.path.join(base_dir, journal)
        journal = journal.split("_")[-1] if "_" in journal else journal
        dir_path, year_dirs, files = next(os.walk(journal_path))
        # year_dirs = [d for d in dirs if len(d) == 4]

        for year in year_dirs:
            year_path = os.path.join(journal_path, year)
            dir_path, month_dirs, files = next(os.walk(year_path))

            for month in month_dirs:
                month_path = os.path.join(year_path, month)
                dir_path, day_dirs, files = next(os.walk(month_path))

                for day in day_dirs:
                    day_path = os.path.join(month_path, day)
                    dir_path, edition_dirs, files = next(os.walk(day_path))

                    for edition in edition_dirs:
                        edition_path = os.path.join(day_path, edition)
                        try:
                            detected_issue = IssueDir(
                                journal,
                                date(int(year), int(month), int(day)),
                                edition,
                                edition_path,
                            )
                            logger.debug(
                                "Found an issue: {}".format(str(detected_issue))
                            )
                            detected_issues.append(detected_issue)
                        except ValueError:
                            logger.warning(
                                "Path {} is not a valid issue directory".format(
                                    edition_path
                                )
                            )
    return detected_issues


def detect_journal_issues(base_dir, newspapers):
    """Parse a directory structure and detect newspaper issues to be imported.

    :param base_dir: the root of the directory structure
    :type base_dir: IssueDir
    :param newspapers: the list of newspapers to consider (acronym blank separated)
    :type newspapers: str
    :return: list of `IssueDir` instances
    :rtype: list
    """
    # newspapers = [journal.split("_")[-1] if "_" in journal else journal for journal in newspapers]

    detected_issues = []
    dir_path, dirs, files = next(os.walk(base_dir))
    journal_dirs = [d for d in dirs if d == newspapers]

    for journal in journal_dirs:
        journal_path = os.path.join(base_dir, journal)
        journal = journal.split("_")[-1] if "_" in journal else journal
        dir_path, year_dirs, files = next(os.walk(journal_path))

        for year in year_dirs:
            if "_" not in year:
                year_path = os.path.join(journal_path, year)
                dir_path, month_dirs, files = next(os.walk(year_path))

                for month in month_dirs:
                    if "_" not in month:
                        month_path = os.path.join(year_path, month)
                        dir_path, day_dirs, files = next(os.walk(month_path))

                        for day in day_dirs:
                            if "_" not in day:
                                day_path = os.path.join(month_path, day)
                                # concerning `edition="a"`: for now, no cases of newspapers
                                # published more than once a day in Olive format (but it
                                # may come later on)
                                detected_issue = IssueDir(
                                    journal,
                                    date(int(year), int(month), int(day)),
                                    "a",
                                    day_path,
                                )
                                logger.debug(
                                    "Found an issue: {}".format(str(detected_issue))
                                )
                                detected_issues.append(detected_issue)
    return detected_issues


def check_filenaming(file_basename):
    """Checks whether a filename complies with our naming convention (GDL-1900-01-10-a-p0001)

    :param file_basename: page file (txt or image)
    :type file_basename: str
    """
    page_pattern = re.compile(r"^[A-Z]+-\d{4}-\d{2}-\d{2}-[a-z]-p\d{4}$")
    return page_pattern.match(file_basename)


def get_issueshortpath(issuedir):
    """Returns short version of issue dir path"""

    path = issuedir.path
    return path[path.index(issuedir.journal) :]



import dask
import os
import pytest
# global variables are imported from conftest.py
from conftest import S3_CANONICAL_BUCKET
from impresso_commons.text.rebuilder import rebuild_issues, compress
from impresso_commons.path.path_s3 import read_s3_issues
from impresso_commons.utils.utils import get_pkg_resource
from dask.distributed import Client
from contextlib import ExitStack
import logging

logger = logging.getLogger(__name__)

DASK_WORKERS_NUMBER = 8
DASK_MEMORY_LIMIT = "2G"

# Use an env var to determine the type of dask scheduling to run:
# 1) synchronous; distributed external or distributed internal
try:
    DASK_SCHEDULER_STRATEGY = os.environ['PYTEST_DASK_SCHEDULER']
except KeyError:
    DASK_SCHEDULER_STRATEGY = 'internal'

if DASK_SCHEDULER_STRATEGY == 'internal':
    client = Client(
        processes=False,
        n_workers=DASK_WORKERS_NUMBER,
        threads_per_worker=1,
        memory_limit=DASK_MEMORY_LIMIT
    )
    print(f"Dask client {client}")
    print(f"Dask client {client.scheduler_info()['services']}")

elif DASK_SCHEDULER_STRATEGY == 'synchronous':
    # it does not work perfectly but almost
    dask.config.set(scheduler="synchronous")
    client = None

elif DASK_SCHEDULER_STRATEGY == 'external':
    client = Client('localhost:8686')
    print(f"Dask client {client}")

limit_issues = 10
test_data = [
    ("NZZ", 1897, limit_issues),
    ("JDG", 1830, limit_issues),
    ("JDG", 1862, limit_issues),
    ("GDL", 1806, limit_issues),
    ("IMP", 1994, limit_issues),
    ("luxzeit1858", 1858, limit_issues),
    ("indeplux", 1905, limit_issues),
    ("luxwort", 1860, limit_issues),
    ("buergerbeamten", 1909, limit_issues),
    ("FedGazDe", 1849, limit_issues),
    ("excelsior", 1911, limit_issues),
    ("oecaen", 1914, limit_issues)

]

@pytest.mark.parametrize("newspaper_id, year, limit", test_data)
def test_rebuild_solr(newspaper_id : str, year : int, limit : int):
    file_mng = ExitStack()
    input_bucket_name = S3_CANONICAL_BUCKET
    outp_dir = get_pkg_resource(file_mng, 'data/rebuilt')

    input_issues = read_s3_issues(newspaper_id, year, input_bucket_name)
    print(f'{newspaper_id}/{year}: {len(input_issues)} issues to rebuild')
    print(f'limiting test rebuild to first {limit} issues.')

    issue_key, json_files = rebuild_issues(
        issues=input_issues[:limit],
        input_bucket=input_bucket_name,
        output_dir=outp_dir,
        dask_client=client,
        format='solr'
    )

    result = compress(issue_key, json_files, outp_dir)
    logger.info(result)
    assert result is not None
    file_mng.close()

def test_rebuild_for_passim():
    input_bucket_name = S3_CANONICAL_BUCKET
    file_mng = ExitStack()
    outp_dir = get_pkg_resource(file_mng, 'data/rebuilt-passim')

    input_issues = read_s3_issues("luxwort", "1848", input_bucket_name)

    issue_key, json_files = rebuild_issues(
        issues=input_issues[:50],
        input_bucket=input_bucket_name,
        output_dir=outp_dir,
        dask_client=client,
        format='passim',
        filter_language=['fr']
    )
    logger.info(f'{issue_key}: {json_files}')
    file_mng.close()


"""Command-line script to generate manifest 0.

Usage:
    rebuilt_manifest_0.py --output-bucekt=<ob> --temp-dir=<td> 

Options:

--output-bucekt=<ob>  Outout s3 bucket for the manifest.
--temp-dir=<td>  Path in local file-system where to write the manifest.
"""

import os
import json
import pandas as pd
import git
from dask import bag as db
from dask import dataframe as dd
from dask.distributed import Client
import numpy as np
from pathlib import Path
from time import strftime
import copy
from typing import Any
from docopt import docopt
from impresso_commons.utils.s3 import (
    fixed_s3fs_glob,
    alternative_read_text,
    upload,
    get_storage_options,
    get_boto3_bucket,
    IMPRESSO_STORAGEOPT,
)
from impresso_commons.versioning.helpers import (
    DataStage,
    read_manifest_from_s3,
    validate_stage,
    clone_git_repo,
    write_and_push_to_git,
    write_dump_to_fs,
)
from impresso_commons.path import parse_canonical_filename
from impresso_commons.path.path_fs import IssueDir
from impresso_commons.path.path_s3 import read_s3_issues, list_newspapers
from impresso_commons.versioning.data_statistics import (
    NewspaperStatistics,
    POSSIBLE_GRANULARITIES,
)
from impresso_commons.versioning.data_manifest import DataManifest
from collections import defaultdict


def chunk(s):
    """
    The function applied to the
    individual partition (map)
    """
    return s.apply(lambda x: list(set(x)))


def agg(s):
    """
    The function whic will aggrgate
    the result from all the partitions(reduce)
    """
    s = s._selected_obj
    return s.groupby(level=list(range(s.index.nlevels))).sum()


def finalize(s):
    """
    The optional functional that will be
    applied to the result of the agg_tu functions
    """
    return s.apply(lambda x: len(set(x)))


# aggregating function implementing np.nunique()
tunique = dd.Aggregation("tunique", chunk, agg, finalize)


def main():
    arguments = docopt(__doc__)
    mft_s3_output_bucket = arguments["--output-bucekt"]
    temp_dir = arguments["--temp-dir"]

    s3_rebuilt_bucket = "rebuilt-data"

    pycommons_repo = git.Repo("/home/piconti/impresso-pycommons")
    # bucket corresponding to the input data of the data currently in 'rebuilt-data'
    mft_s3_input_bucket = "canonical-data"

    rebuilt_mft_0 = DataManifest(
        data_stage="rebuilt",  # DataStage.REBUILT also works
        s3_output_bucket=mft_s3_output_bucket,
        s3_input_bucket=mft_s3_input_bucket,
        git_repo=pycommons_repo,
        temp_dir=temp_dir,
        staging=True,
    )

    client = Client(n_workers=16, threads_per_worker=2)
    print(client)

    print(f"Fetching the rebuilt files from the bucket s3://{s3_rebuilt_bucket}")

    rebuilt_files = fixed_s3fs_glob(os.path.join(s3_rebuilt_bucket, "*.jsonl.bz2"))

    print(f"Reading the contents of the {len(rebuilt_files)} fetched files...")
    # lazy object, to comput once reduced
    rebuilt_bag = db.read_text(rebuilt_files, storage_options=IMPRESSO_STORAGEOPT).map(
        json.loads
    )

    # for each rebuilt content-item, fetch the information necessary to compute the statistics we seek
    rebuilt_count_df = (
        rebuilt_bag.map(
            lambda ci: {
                "np_id": ci["id"].split("-")[0],
                "year": ci["id"].split("-")[1],
                "issue_id": "-".join(
                    ci["id"].split("-")[:-1]
                ),  # count the issues represented
                "n_content_items": 1,  # each element of the bag corresponds to one content-item
                "n_tokens": (
                    len(ci["ft"].split()) if "ft" in ci else 0
                ),  # split on spaces to count tokens
            }
        )
        .to_dataframe(
            meta={
                "np_id": str,
                "year": str,
                "issue_id": str,
                "n_content_items": int,
                "n_tokens": int,
            }
        )
        .persist()
    )

    # agggregate them at the scale of the entire corpus
    # first groupby title, year and issue to also count the individual issues present
    aggregated_df = (
        rebuilt_count_df
        # .groupby(by=['np_id', 'year', 'issue_id'])
        # .agg({'n_content_items': sum, 'n_tokens': sum})
        # .reset_index()
        # .groupby(by=['np_id', 'year'])
        # .agg({'issue_id': 'count', 'n_content_items': sum, 'n_tokens': sum})
        .groupby(by=["np_id", "year"])
        .agg({"issue_id": tunique, "n_content_items": sum, "n_tokens": sum})
        .rename(
            columns={
                "issue_id": "issues",
                "n_content_items": "content_items_out",
                "n_tokens": "ft_tokens",
            }
        )
        .reset_index()
    )

    print("Obtaining the yearly rebuilt statistics for the entire corpus")
    # return as a list of dicts
    stats_as_dict = aggregated_df.to_bag(format="dict").compute()

    rebuilt_stats = copy.deepcopy(stats_as_dict)

    print("Populating the manifest with the resulting yearly statistics...")
    # populate the manifest with these statistics
    for stats in rebuilt_stats:
        title = stats["np_id"]
        year = stats["year"]
        del stats["np_id"]
        del stats["year"]
        rebuilt_mft_0.add_by_title_year(title, year, stats)

    print("Finalizing the manifest, and computing the result...")

    note = f"Initial Manifest computed retroactively on the rebuilt data present in the bucket s3://{s3_rebuilt_bucket}."
    rebuilt_mft_0.append_to_notes(note)
    rebuilt_mft_0.compute(export_to_git_and_s3=False)
    rebuilt_mft_0.validate_and_export_manifest(push_to_git=False)


if __name__ == "__main__":
    main()



"""
Simple CLI script to delete versioned keys from S3

Usage:
    impresso_commons/utils/s3_delete.py --bucket=<b> --prefix=<p>

Options:
    --bucket=<b>    Target S3 bucket
    --prefix=<p>    Prefix of keys to delete
"""

from docopt import docopt

from impresso_commons.utils.s3 import get_s3_resource
from impresso_commons.utils import user_confirmation

def delete_versioned_keys(
    client,
    Bucket,
    Prefix,
    IsTruncated=True,
    MaxKeys=1000,
    KeyMarker=None
):
    """TODO"""
    while IsTruncated == True:
            if not KeyMarker:
                version_list = client.list_object_versions(
                        Bucket=Bucket,
                        MaxKeys=MaxKeys,
                        Prefix=Prefix)
            else:
                version_list = client.list_object_versions(
                        Bucket=Bucket,
                        MaxKeys=MaxKeys,
                        Prefix=Prefix,
                        KeyMarker=KeyMarker)

            try:
                objects = []
                versions = version_list['Versions']
                for v in versions:
                        objects.append({
                            'VersionId':v['VersionId'],
                            'Key': v['Key']
                        })
                response = client.delete_objects(
                    Bucket=Bucket,
                    Delete={'Objects':objects}
                )
                print(f"Deleted {len(response['Deleted'])} keys")
            except:
                pass

            try:
                objects = []
                delete_markers = version_list['DeleteMarkers']
                for d in delete_markers:
                        objects.append({
                            'VersionId':d['VersionId'],
                            'Key': d['Key']
                        })
                response = client.delete_objects(
                    Bucket=Bucket,
                    Delete={'Objects':objects}
                )
                print(f"Deleted {len(response['Deleted'])} keys")
            except:
                pass

            IsTruncated = version_list['IsTruncated']
            try:
                KeyMarker = version_list['NextKeyMarker']
            except KeyError:
                print("Done!")


def main():
    args = docopt(__doc__)
    b = args['--bucket']
    p = args['--prefix']
    q_1 = f"all keys with prefix `{p}`"
    q = f"\nAre you sure you want to delete {q_1} from bucket `s3://{b}` ?"

    if user_confirmation(question=q):
        print("Ok, let's start (it will take a while!)")
        s3_client = get_s3_resource().meta.client
        delete_versioned_keys(client=s3_client, Bucket=b, Prefix=p)
    else:
        print('Ok then, see ya!')

if __name__ == '__main__':
    main()



"""Command-line script to generate manifest 0.

Usage:
    manifest_0.py --canonical-bucket=<cb> --manifest-out-path=<mop> 

Options:

--canonical-bucket=<cb>  S3 bucket to use to compute the statistics and to upload the manifest to.
--manifest-out-path=<mop>  Path in local file-system where to write the manifest.
"""

import json
from docopt import docopt
from time import strftime
import copy
from typing import Any

import dask.bag as db

from impresso_commons.utils.s3 import (
    upload,
)
from impresso_commons.path.path_s3 import fetch_files
from impresso_commons.versioning.data_statistics import (
    NewspaperStatistics,
)


# adapted from https://github.com/impresso/impresso-data-sanitycheck/blob/master/sanity_check/contents/stats.py#L241
def compute_canonical_stats_for_manifest(
    s3_canonical_issues: db.core.Bag,
) -> list[dict[str, Any]]:
    """Computes some statistics on issues and pages in provided issue files

    Args:
        s3_canonical_issues (db.core.Bag): Dask.bag containing the issues to compute
            the statistics on.

    Returns:
        list[dict[str, Any]]: Canonical statistics as a list of dicts.
    """

    print("Fetched all issues, gathering desired information.")
    pages_count_df = (
        s3_canonical_issues.map(
            lambda i: {
                "np_id": i["id"].split("-")[0],
                "year": i["id"].split("-")[1],
                "id": i["id"],
                "issue_id": i["id"],
                "n_pages": len(set(i["pp"])),
                "n_content_items": len(i["i"]),
                "n_images": len(
                    [item for item in i["i"] if item["m"]["tp"] == "image"]
                ),
            }
        )
        .to_dataframe(
            meta={
                "np_id": str,
                "year": str,
                "id": str,
                "issue_id": str,
                "n_pages": int,
                "n_images": int,
                "n_content_items": int,
            }
        )
        .set_index("id")
        .persist()
    )

    # cum the counts for all values collected
    aggregated_df = (
        pages_count_df.groupby(by=["np_id", "year"])
        .agg(
            {
                "n_pages": sum,
                "issue_id": "count",
                "n_content_items": sum,
                "n_images": sum,
            }
        )
        .rename(
            columns={
                "issue_id": "issues",
                "n_pages": "pages",
                "n_content_items": "content_items_out",
                "n_images": "images",
            }
        )
        .reset_index()
    )

    print("Finished grouping and aggregating by title and year.")
    # return as a list of dicts
    return aggregated_df.to_bag(format="dict").compute()


def pretty_print(np_stats: NewspaperStatistics) -> dict[str, Any]:
    stats_dict = {
        "stage": np_stats.format.value,
        "granularity": np_stats.granularity,
        "element": np_stats.element,
        "nps_stats": {k: v for k, v in np_stats.counts.items() if v > 0},
    }
    if stats_dict["granularity"] == "corpus":
        del stats_dict["element"]
    return stats_dict


def new_np_stats(stats: dict[str, int]) -> NewspaperStatistics:
    element = f"{stats['np_id']}-{stats['year']}"
    del stats["np_id"]
    del stats["year"]
    return NewspaperStatistics("canonical", "year", element, counts=stats)


def new_media(stats: dict[str, int], manifest_data: dict[str, Any]) -> dict[str, Any]:
    title = stats["np_id"]
    print(f"Creating new media dict for {title}.")
    new_media_stats = new_np_stats(stats)
    return {
        "media_title": title,
        "last_modification_date": manifest_data["mft_generation_date"],
        "update_type": None,
        "update_level": "title",
        "update_targets": [],
        "code_git_commit": None,
        "media_statistics": [new_media_stats],
    }


def aggregate_stats_for_title(title: str, media_dict: dict[str, Any]):
    print(f"Inside aggregate_stats_for_title for {title}.")
    # instantiate a NewspaperStatistics object for the title
    title_cumm_stats = NewspaperStatistics("canonical", "title", title)
    # instantiate the list of counts for display
    pretty_counts = []
    for np_year_stat in media_dict["media_statistics"]:
        # add the title yearly counts
        title_cumm_stats.add_counts(np_year_stat.counts)
        pretty_counts.append(pretty_print(np_year_stat))
    # insert the title-level statistics at the top of the statistics
    pretty_counts.insert(0, pretty_print(title_cumm_stats))
    media_dict["media_statistics"] = pretty_counts

    return media_dict, title_cumm_stats


def main():
    arguments = docopt(__doc__)
    s3_canonical_bucket = arguments["--canonical-bucket"]
    manifest_out_path = arguments["--manifest-out-path"]

    version = "v0.0.1"
    manifest_out_name = f"canonical_{version.replace('.','-')}.json"

    print(
        f"Generating manifest {manifest_out_name} using the s3 bucket {s3_canonical_bucket}."
    )

    manifest_data = {
        "mft_version": version,
        "mft_generation_date": strftime("%Y-%m-%d %H:%M:%S"),
        "mft_s3_path": f"s3://{s3_canonical_bucket}/{manifest_out_name}",
        "input_mft_s3_path": None,
        "input_mft_git_path": None,
        "code_git_commit": None,
        "media_list": [],
        "overall_statistics": [],
        "notes": f"Initial Manifest computed retroactively on the canonical data present in the bucket s3://{s3_canonical_bucket}.",
    }

    s3_canonical_issues, _ = fetch_files(s3_canonical_bucket, compute=False)

    canonical_stats = compute_canonical_stats_for_manifest(s3_canonical_issues)
    canonical_counts = copy.deepcopy(canonical_stats)

    print(
        "Finished gathering and aggregating all counts. Start populating the manifests."
    )

    # instantiate the media list with NewspaperStatistics ojects
    canonical_media_list = {}
    for d in canonical_counts:
        title = d["np_id"]
        if title in canonical_media_list:
            # if the title is already present in the media list, append new statistics
            canonical_media_list[title]["media_statistics"].append(new_np_stats(d))
        else:
            # if the title is not in the media list, initialize a new "media" object
            canonical_media_list[title] = new_media(d, manifest_data)

    # aggregate the statistics per title
    full_title_stats = []
    for title, media in canonical_media_list.items():
        media_dict, title_cumm_stats = aggregate_stats_for_title(title, media)
        # update the canonical_media_list with the new media_dict
        canonical_media_list[title] = media_dict
        # save the title level statistics for the overall statistics
        full_title_stats.append(title_cumm_stats)

    # set the final media list (only the values)
    manifest_data["media_list"] = list(canonical_media_list.values())

    print("Now computing the overall statistics")
    # compute the overall statistics
    corpus_stats = NewspaperStatistics("canonical", "corpus", "")
    for np_stats in full_title_stats:
        corpus_stats.add_counts(np_stats.counts)
    # add the number of titles present in corpus
    corpus_stats.add_counts({"titles": len(full_title_stats)})
    # set the final overall counts
    manifest_data["overall_statistics"] = [pretty_print(corpus_stats)]

    f_path = f"{manifest_out_path}/{manifest_out_name}"

    print(f"Finished creating the manifest. Writing to fs and S3: {f_path}.")

    with open(f_path, "w", encoding="utf-8") as f:
        json.dump(manifest_data, f, ensure_ascii=False, indent=4)

    # upload to s3
    upload(f_path, bucket_name=s3_canonical_bucket)


if __name__ == "__main__":
    main()



"""Code for parsing impresso's S3 directory structures."""

import os
import json
import logging
import warnings
from datetime import date
from collections import namedtuple
from typing import Optional, Union

from dask.diagnostics import ProgressBar
import dask.bag as db

from impresso_commons.path import id2IssueDir
from impresso_commons.utils.s3 import get_s3_client, get_s3_versions, fixed_s3fs_glob
from impresso_commons.utils.s3 import IMPRESSO_STORAGEOPT

logger = logging.getLogger(__name__)
_WARNED = False
# a simple data structure to represent input directories
IssueDir = namedtuple("IssueDirectory", ["journal", "date", "edition", "path"])


# a data structure to represent content items (articles or pages)
class s3ContentItem:
    def __init__(
        self,
        journal,
        date,
        edition,
        number,
        key_name,
        doc_type=None,
        rebuilt_version=None,
        canonical_version=None,
    ):
        self.journal = journal
        self.date = date
        self.edition = edition
        self.number = number
        self.doc_type = doc_type
        self.key_name = key_name
        self.rebuilt_version = rebuilt_version
        self.canonical_version = canonical_version


def _list_bucket_paginator(bucket_name, prefix="", accept_key=lambda k: True):
    """
    List the content of a bucket using pagination.
    No filtering besides indicated prefix and accept_key lambda.

    :param bucket_name: string, e.g. 'original-canonical-data'
    :param prefix: string, e.g. 'GDL/1950' - refers to the pseudo hierarchical
        structure within the bucket
    :param accept_key: lambda function, to accept or reject a specific key
    @return: arrays of keys
    """
    client = get_s3_client()
    paginator = client.get_paginator("list_objects")
    page_iterator = paginator.paginate(Bucket=bucket_name, Prefix=prefix)
    keys = []
    for page in page_iterator:
        if "Contents" in page:
            for key in page["Contents"]:
                keyString = key["Key"]
                if accept_key(keyString):
                    keys.append(keyString)
    return keys if keys else []


def _list_bucket_paginator_filter(
    bucket_name, prefix="", accept_key=lambda k: True, config=None
):
    """
    List the content of a bucket using pagination, with a filter.
    :param bucket_name: string, e.g. 'original-canonical-data'
    :param prefix: string, e.g. 'GDL/1950' - refers to the pseudo hierarchical structure within the bucket
    :param accept_key: lambda function, to accept or reject a specific key
    :param config: a dict with newspaper acronyms as keys and array of year interval as values:
    e.g. { "GDL": [1950, 1960], "JDG": [1890, 1900] }. Last year is excluded.
    @return: arrays of keys
    """
    filtered_keys = []

    # building a list of prefixes from the config information
    for np in config:
        # if years are specified, take the range
        if config[np]:
            prefixes = [
                np + "/" + str(item) for item in range(config[np][0], config[np][1])
            ]
        # otherwise prefix is just the newspaper
        else:
            prefixes = [np]

        # retrieving keys using the prefixes
        print(f"Detecting items for {np} for years {prefixes}")
        for prefix in prefixes:
            client = get_s3_client()
            paginator = client.get_paginator("list_objects")
            page_iterator = paginator.paginate(Bucket=bucket_name, Prefix=prefix)
            for page in page_iterator:
                if "Contents" in page:
                    for key in page["Contents"]:
                        keyString = key["Key"]
                        if accept_key(keyString):
                            filtered_keys.append(keyString)
    return filtered_keys if filtered_keys else []


def _key_to_issue(key_info):
    """Instantiate an IssueDir from a key info tuple.
    :param key_info: tuple (key_name, key_versionid, date_lastupdated)
    @return: IssueDir
    """
    key = key_info[0]
    name_no_prefix = key.split("/")[-1]
    canon_name = name_no_prefix.replace("-issue.json", "")
    journal, year, month, day, edition = canon_name.split("-")
    path = key
    issue = IssueDir(journal, date(int(year), int(month), int(day)), edition, path)
    return issue._asdict()


def _key_to_contentitem(key_info):
    """
    Instantiate an ContentItem from a key info tuple.
    :param key_info: tuple (key_name, key_versionid, date_lastupdated)
    @return: ContentItem
    """
    key = key_info[0]  # GDL/1950/01/06/a/GDL-1950-01-06-a-i0056.json
    name_no_prefix = key.split("/")[-1]
    canon_name = name_no_prefix.replace(".json", "")
    journal, year, month, day, edition, number = canon_name.split("-")
    ci_type = number[:1]
    path = key
    return s3ContentItem(
        journal,
        date(int(year), int(month), int(day)),
        edition,
        number[1:],
        path,
        ci_type,
        rebuilt_version=key_info[1],
    )


def _process_keys(key_name, bucket_name, item_type):
    """
    Convert a key in an impresso object: IssueDir or ContentItem
    :param key_name:
    :param bucket_name:
    :param item_type:
    @return:
    """
    # choose the type of build to use
    build = _key_to_issue if item_type == "issue" else _key_to_contentitem
    # retrieve versioning information
    version_id, last_modified = get_s3_versions(bucket_name, key_name)[0]
    # return the object, using the appropriate function
    return build((key_name, version_id, last_modified))


def impresso_iter_bucket(
    bucket_name, item_type=None, prefix=None, filter_config=None, partition_size=15
):
    """
    Iterate over a bucket, possibly with a filter, and return an array of either IssueDir or ContentItem.
    VALID ONLY for original-canonical data, where there is individual files for issues and content items (articles).
    :param bucket_name: string, e.g. 'original-canonical-data'
    :param item_type: 'issue' or 'item'
    :param prefix: string, e.g. 'GDL/1950', used to filter key. Exclusive of 'filter_config'
    :param filter_config: a dict with newspaper acronyms as keys and array of year interval as values:
    e.g. { "GDL": [1950, 1960], "JDG": [1890, 1900] }. Last year is excluded.
    :param partition_size: partition size of dask to build the object (Issuedir or ContentItem)
    @return: an array of (filtered) IssueDir or ContentItems.
    """
    global _WARNED
    if not _WARNED:
        warning = (
            "This function is depreciated and cannot be trusted to yield"
            " correct outputs. Please use s3_iter_bucket instead."
        )
        logger.warning(warning)
        warnings.warn(warning, DeprecationWarning)
    # either prefix or config, but not both
    if prefix and filter_config:
        logger.error("Provide either a prefix or a config but not both")
        return None

    # check which kind of object to build, issue or content_item
    suffix = "issue.json" if item_type == "issue" else ".json"

    # collect keys using pagination
    logger.info("Start collecting key from s3 (not parallel)")
    if filter_config is None:
        keys = _list_bucket_paginator(
            bucket_name, prefix, accept_key=lambda key: key.endswith(suffix)
        )
    else:
        keys = _list_bucket_paginator_filter(
            bucket_name,
            accept_key=lambda key: key.endswith(suffix),
            config=filter_config,
        )

    # build IssueDir or ContentItem from the keys, using dask.
    logger.info(f"Start processing key.")
    ci_bag = db.from_sequence(
        keys, partition_size
    )  # default partition_size in dask: about 100
    ci_bag = ci_bag.map(_process_keys, bucket_name=bucket_name, item_type=item_type)
    with ProgressBar():
        result = ci_bag.compute()

    return result


def s3_iter_bucket(bucket_name, prefix, suffix):
    """
    Iterate over a bucket and return all keys with `prefix` and `suffix`.

    >>> b = get_bucket("myBucket", create=False)
    >>> k = s3_iter_bucket(b.name, prefix='GDL', suffix=".bz2")
    >>>
    :param bucket_name: the name of the bucket
    :type bucket_name: str
    :param prefix: beginning of the key
    :type prefix: str
    :param key_suffix: how the key ends
    :type prefix: str
    @return: array of keys
    """
    return _list_bucket_paginator(
        bucket_name, prefix, accept_key=lambda key: key.endswith(suffix)
    )


def s3_filter_archives(bucket_name, config, suffix=".jsonl.bz2"):
    """
    Iterate over bucket and filter according to config and suffix.
    Config is a dict where k= newspaper acronym and v = array of 2 years, considered as time interval.
    Example:
        config = {
            "GDL" : [1960, 1970], => will take all years in interval
            "JDG": [], => Empty array means no filter, all years.
            "GDL": [1798, 1999, 10] => take each 10th item within sequence of years

        }

    :param bucket_name: the name of the bucket
    :type bucket_name: str
    :param config: newspaper/years to consider
    :type config: Dict
    :param key_suffix: end of the key
    :type prefix: str
    :return: array of keys
    """
    filtered_keys = []
    accept_key = lambda k: True
    keynames = ["allyears"]  # todo

    # generate keynames from config (e.g. 'GDL/GDL-1950.jsonl.bz2')
    for np in config:
        if config[np]:
            tmp = [
                np + "/" + np + "-" + str(item) + suffix
                for item in range(config[np][0], config[np][1])
            ]
            if len(config[np]) == 2:
                keynames = tmp
            elif len(config[np]) == 3:
                keynames = tmp[:: config[np][2]]

            accept_key = lambda k: k in keynames

        # retrieving keys
        print(f"Detecting items for {np} with suffixes {keynames}")
        client = get_s3_client()
        paginator = client.get_paginator("list_objects")
        page_iterator = paginator.paginate(Bucket=bucket_name, Prefix=np)
        for page in page_iterator:
            if "Contents" in page:
                for key in page["Contents"]:
                    keyString = key["Key"]
                    if accept_key(keyString):
                        filtered_keys.append(keyString)

    return filtered_keys if filtered_keys else []


def read_s3_issues(newspaper, year, input_bucket):

    def add_version(issue):
        issue["s3_version"] = None
        return issue

    issue_path_on_s3 = (
        f"{input_bucket}/{newspaper}/issues/{newspaper}-{year}-issues.jsonl.bz2"
    )
    issues = (
        db.read_text(issue_path_on_s3, storage_options=IMPRESSO_STORAGEOPT)
        .map(json.loads)
        .map(add_version)
        .map(lambda x: (id2IssueDir(x["id"], issue_path_on_s3), x))
        .compute()
    )
    return issues


def list_newspapers(
    bucket_name: str,
    s3_client=get_s3_client(),
    page_size: int = 10000,
) -> list[str]:
    """List newspapers contained in an s3 bucket with impresso data.

    Note:
        25,000 seems to be the maximum `PageSize` value supported by
        SwitchEngines' S3 implementation (ceph).
    Note:
        Copied from https://github.com/impresso/impresso-data-sanitycheck/tree/master/sanity_check/contents/s3_data.py

    Args:
        bucket_name (str): Name of the S3 bucket to consider
        s3_client (optional): S3 client to use. Defaults to get_s3_client().
        page_size (int, optional): Pagination configuration. Defaults to 10000.

    Returns:
        list[str]: List of newspaper (aliases) present in the given S3 bucket.
    """
    print(f"Fetching list of newspapers from {bucket_name}")

    if "s3://" in bucket_name:
        bucket_name = bucket_name.replace("s3://", "").split("/")[0]

    paginator = s3_client.get_paginator("list_objects")

    newspapers = set()
    for n, resp in enumerate(
        paginator.paginate(Bucket=bucket_name, PaginationConfig={"PageSize": page_size})
    ):
        # means the bucket is empty
        if "Contents" not in resp:
            continue

        for f in resp["Contents"]:
            newspapers.add(f["Key"].split("/")[0])
        msg = (
            f"Paginated listing of keys in {bucket_name}: page {n + 1}, listed "
            f"{len(resp['Contents'])}"
        )
        logger.info(msg)

    print(f"{bucket_name} contains {len(newspapers)} newspapers")

    return newspapers


def list_files(
    bucket_name: str,
    file_type: str = "issues",
    newspapers_filter: Optional[list[str]] = None,
) -> tuple[Optional[list[str]], Optional[list[str]]]:
    """List the canonical files located in a given S3 bucket.

    Note:
        adapted from https://github.com/impresso/impresso-data-sanitycheck/tree/master/sanity_check/contents/s3_data.py

    Args:
        bucket_name (str): S3 bucket name.
        file_type (str, optional): Type of files to list, possible values are "issues",
            "pages" and "both". Defaults to "issues".
        newspapers_filter (list[str] | None, optional): List of newspapers to consider.
            If None, all will be considered. Defaults to None.

    Raises:
        NotImplementedError: The given `file_type` is not one of ['issues', 'pages', 'both'].

    Returns:
        tuple[list[str] | None, list[str] | None]: [0] List of issue files or None and
            [1] List of page files or None based on `file_type`
    """
    if file_type not in ["issues", "pages", "both"]:
        logger.error("The provided type is not one of ['issues', 'pages', 'both']!")
        raise NotImplementedError

    # initialize the output lists
    issue_files, page_files = None, None
    # list the newspapers in the bucket
    newspapers = list_newspapers(bucket_name)

    if newspapers_filter is not None:
        suffix = f"for the provided newspapers {newspapers_filter}"
    else:
        suffix = ""

    if file_type in ["issues", "both"]:
        issue_files = [
            file
            for np in newspapers
            if newspapers_filter is not None and np in newspapers_filter
            for file in fixed_s3fs_glob(os.path.join(bucket_name, f"{np}/issues/*"))
        ]
        print(f"{bucket_name} contains {len(issue_files)} .bz2 issue files {suffix}")
    if file_type in ["pages", "both"]:
        page_files = [
            file
            for np in newspapers
            if newspapers_filter is not None and np in newspapers_filter
            for file in fixed_s3fs_glob(f"{os.path.join(bucket_name, f'{np}/pages/*')}")
        ]
        print(f"{bucket_name} contains {len(page_files)} .bz2 page files {suffix}")

    return issue_files, page_files


def fetch_files(
    bucket_name: str,
    compute: bool = True,
    file_type: str = "issues",
    newspapers_filter: Optional[list[str]] = None,
) -> tuple[Union[db.core.Bag, list[str], None], Union[db.core.Bag, list[str], None]]:
    """Fetch issue and/or page canonical JSON files from an s3 bucket.

    If compute=True, the output will be a list of the contents of all files in the
    bucket for the specified newspapers and type of files.
    If compute=False, the output will remain in a distributed dask.bag.

    Based on file_type, the issue files, page files or both will be returned.
    In the returned tuple, issues are always in the first element and pages in the
    second, hence if file_type is not 'both', the tuple entry corresponding to the
    undesired type of files will be None.

    Note:
        adapted from https://github.com/impresso/impresso-data-sanitycheck/tree/master/sanity_check/contents/s3_data.py

    Args:
        bucket_name (str): Name of the s3 bucket to fetch the files form
        compute (bool, optional): Whether to compute result and output as list.
            Defaults to True.
        file_type: (str, optional): Type of files to list, possible values are "issues",
            "pages" and "both". Defaults to "issues".
        newspapers_filter: (list[str]|None,optional): List of newspapers to consider.
            If None, all will be considered. Defaults to None.

    Raises:
        NotImplementedError: The given `file_type` is not one of ['issues', 'pages', 'both'].

    Returns:
        tuple[db.core.Bag|None, db.core.Bag|None] | tuple[list[str]|None, list[str]|None]:
            [0] Issue files' contents or None and
            [1] Page files' contents or None based on `file_type`
    """
    if file_type not in ["issues", "pages", "both"]:
        logger.error("The provided type is not one of ['issues', 'pages', 'both']!")
        raise NotImplementedError

    issue_files, page_files = list_files(bucket_name, file_type, newspapers_filter)
    # initialize the outputs
    issue_bag, page_bag = None, None

    msg = "Fetching "
    if issue_files is not None:
        msg = f"{msg} issue ids from {len(issue_files)} .bz2 files, "
        issue_bag = db.read_text(issue_files, storage_options=IMPRESSO_STORAGEOPT).map(
            json.loads
        )
    if page_files is not None:
        # make sure all files are .bz2 files and exactly have the naming format they should
        prev_len = len(page_files)
        page_files = [
            p for p in page_files if ".jsonl.bz2" in p and len(p.split("-")) > 5
        ]
        msg = f"{msg} page ids from {len(page_files)} .bz2 files ({prev_len} files before filtering), "
        page_bag = db.read_text(page_files, storage_options=IMPRESSO_STORAGEOPT).map(
            json.loads
        )

    logger.info(msg)

    if compute:
        page_bag = page_bag.compute() if page_files is not None else page_bag
        issue_bag = issue_bag.compute() if issue_files is not None else issue_bag

    return issue_bag, page_bag



#!/usr/bin/env python3
# coding: utf-8
import pytest
from impresso_commons.path.path_s3 import (impresso_iter_bucket,
                                           s3_filter_archives, s3_iter_bucket)
from impresso_commons.utils.s3 import get_bucket

impresso_iter_bucket_testdata = [
    #("original-canonical-data", "issue", None, 'GDL/issues/GDL-1950', 'not None'), 
    ("rebuilt-data", "item", {"GDL": [1950, 1951]}, None, 'Empty'),
    ("original-canonical-data", "item", {"GDL": [1950, 1951]}, 'GDL/pages/GDL-1950', 'None'),
]

@pytest.mark.parametrize(
    "bucket,type,np_config,prefix,expected", 
    impresso_iter_bucket_testdata
)
def test_impresso_iter_bucket(bucket, type, np_config, prefix, expected):

    iter_items = impresso_iter_bucket(
        bucket_name=bucket,
        item_type=type,
        prefix = prefix,
        filter_config=np_config
    )

    if expected == 'not None':
        assert iter_items is not None
        assert len(iter_items) > 0
    elif expected == 'Empty':
        assert len(iter_items) == 0
    else:
        assert iter_items is None


def test_s3_iter_bucket():
    b = get_bucket("canonical-rebuilt", create=False)
    keys = s3_iter_bucket(b.name, prefix='GDL', suffix=".bz2")
    assert keys is not None
    assert len(keys) > 0


def test_s3_filter_archives():
    config = {"GDL": [1950, 1959]}
    b = get_bucket("canonical-rebuilt", create=False)
    keys = s3_filter_archives(b.name, config=config)
    assert keys is not None
    assert len(keys) == 9


def test_s3_filter_archives_timebucket():
    config = {"GDL": [1799, 1999, 10]}
    b = get_bucket("canonical-rebuilt", create=False)
    keys = s3_filter_archives(b.name, config=config)
    assert keys is not None
    assert len(keys) == 20



"""Command-line script to generate a manifest for an S3 bucket or partition after a processing.

Usage:
    compute_manifest.py --config-file=<cf> --log-file=<lf> [--scheduler=<sch> --nworkers=<nw> --verbose]

Options:

--config-file=<cf>  Path to configuration json file containing all necessary arguments for the computation of the manifest.
--log-file=<lf>  Path to log file to use.
--scheduler=<sch>  Tell dask to use an existing scheduler (otherwise it'll create one)
--nworkers=<nw>  number of workers for (local) Dask client.
--verbose  Set logging level to DEBUG (by default is INFO).
"""

import json
import os
import traceback
import logging
from typing import Any, Optional
import git
from docopt import docopt

import dask.bag as db
from dask.distributed import Client
from impresso_commons.utils.s3 import fixed_s3fs_glob, IMPRESSO_STORAGEOPT
from impresso_commons.utils.utils import init_logger
from impresso_commons.versioning.helpers import (
    validate_stage,
    DataStage,
    compute_stats_in_canonical_bag,
    compute_stats_in_rebuilt_bag,
    compute_stats_in_entities_bag,
    compute_stats_in_langident_bag,
)
from impresso_commons.versioning.data_manifest import DataManifest

logger = logging.getLogger(__name__)

# list of optional configurations
OPT_CONFIG_KEYS = [
    "input_bucket",
    "newspapers",
    "previous_mft_s3_path",
    "is_patch",
    "patched_fields",
    "only_counting",
    "push_to_git",
    "notes",
]
# list of requirec configurations
REQ_CONFIG_KEYS = [
    "data_stage",
    "output_bucket",
    "git_repository",
    "is_staging",
    "file_extensions",
]


def extract_np_key(s3_key: str, bucket: str) -> str:
    """Extract the newspaper an s3:key corresponds to given the bucket and partition

    eg. s3_key is in format:
    - s3_key: 's3://31-passim-rebuilt-staging/passim/indeplux/indeplux-1889.jsonl.bz2'
    - bucket: '31-passim-rebuilt-staging/passim'
    --> returns 'indeplux'

    Args:
        s3_key (str): Full S3 path of a file (as returned by fixed_s3fs_glob).
        bucket (str): S3 bucket, including partition, in which the newspaper dirs are.

    Returns:
        str: Name of the corresponding newspaper, extracted form the s3 path.
    """
    # in format: 's3://31-passim-rebuilt-staging/passim/indeplux/indeplux-1889.jsonl.bz2'
    key_no_bucket = s3_key.replace(f"s3://{bucket}/", "")
    # Not all buckets separate the data per title, but the title will always come first.
    if "/" in key_no_bucket:
        return key_no_bucket.split("/")[0]
    else:
        return key_no_bucket.split("-")[0]


def get_files_to_consider(config: dict[str, Any]) -> Optional[dict[str, list[str]]]:
    """Get the list of S3 files to consider based on the provided configuration.

    Args:
        config (dict[str, Any]): Configuration parameters with the s3 bucket, titles,
            and file extensions

    Returns:
        dict[str, list[str]] | None: Dict mapping each newspaper to the s3 files to
            consider, or None if no files found.

    Raises:
        ValueError: If `file_extensions` in the config is empty or None.
    """
    if config["file_extensions"] == "" or config["file_extensions"] is None:
        raise ValueError("Config file's `file_extensions` should not be empty or None.")

    ext = config["file_extensions"]
    extension_filter = f"*{ext}" if "." in ext else f"*.{ext}"

    # if newspapers is empty, include all newspapers
    if config["newspapers"] is None or len(config["newspapers"]) == 0:
        logger.info("Fetching the files to consider for all titles...")
        # TODO update list_newspapers to include possibility of partition, and unify both cases
        # return all filenames in the given bucket partition with the correct extension
        files = fixed_s3fs_glob(os.path.join(config["output_bucket"], extension_filter))
        s3_files = {}
        for s3_key in files:
            np = extract_np_key(s3_key, config["output_bucket"])
            if np in s3_files:
                s3_files[np].append(s3_key)
            else:
                s3_files[np] = [s3_key]
        return s3_files

    # here list newspapers instead and s3_files becomes a dict np -> liest of files
    logger.info("Fetching the files to consider for titles %s...", config["newspapers"])
    s3_files = {}
    for np in config["newspapers"]:
        s3_files[np] = fixed_s3fs_glob(
            os.path.join(config["output_bucket"], np, extension_filter)
        )

    return s3_files


def compute_stats_for_stage(
    files_bag: db.core.Bag, stage: DataStage, client: Optional[Client] = None
) -> Optional[list[dict]]:
    """Compute statistics for a specific data stage.

    Args:
        files_bag (db.core.Bag): A bag containing files for statistics computation.
        stage (DataStage): The data stage for which statistics are computed.
        client (Client | None, optional): Dask client to use.

    Returns:
        list[dict] | None]: List of computed yearly statistics, or None if statistics
            computation for the given stage is not implemented.
    """
    match stage:
        case DataStage.CANONICAL:
            return compute_stats_in_canonical_bag(files_bag, client=client)
        case DataStage.REBUILT:
            return compute_stats_in_rebuilt_bag(
                files_bag, include_np=True, client=client
            )
        case DataStage.ENTITIES:
            return compute_stats_in_entities_bag(files_bag, client=client)
        case DataStage.PASSIM:
            return compute_stats_in_rebuilt_bag(
                files_bag, include_np=True, passim=True, client=client
            )
        case DataStage.LANGIDENT:
            return compute_stats_in_langident_bag(files_bag, client=client)
    raise NotImplementedError(
        "The function computing statistics for this DataStage is not yet implemented."
    )


def validate_config(config: dict[str, Any]) -> dict[str, Any]:
    """Ensure all required configurations are defined, add any missing optional ones.

    Args:
        config (dict[str, Any]): Provided configuration dict to compute the manifest.

    Raises:
        ValueError: Some required arguments of the configuration are missing.

    Returns:
        dict[str, Any]: Updated config, with any mssing optional argument set to None.
    """
    logger.info(
        "Validating that the provided configuration has all required arugments."
    )
    if not all([k in config for k in REQ_CONFIG_KEYS]):
        raise ValueError(f"Missing some required configurations: {REQ_CONFIG_KEYS}")

    for k in OPT_CONFIG_KEYS:
        if k not in config:
            logger.debug("%s was missing form the configuration, setting it to None", k)
            config[k] = None

    return config


def create_manifest(
    config_dict: dict[str, Any], client: Optional[Client] = None
) -> None:
    """Given its configuration, generate the manifest for a given s3 bucket partition.

    TODO: add option to agg for all titles together if desired
    TODO: add iptions to exclude NP for all agg types
    TODO: separate further into functions

    Note:
        The contents of the configuration file (or dict) are given in markdown file
        `impresso_commons/data/manifest_config/manifest.config.example.md``

    Args:
        config_dict (dict[str, Any]): Configuration following the guidelines.
        client (Client | None, optional): Dask client to use.
    """
    # if the logger was not previously inialized, do it
    if not logger.hasHandlers():
        init_logger()

    # ensure basic validity of the provided configuration
    config_dict = validate_config(config_dict)
    stage = validate_stage(config_dict["data_stage"])
    logger.info("Provided config validated.")

    logger.info("Starting to generate the manifest for DataStage: '%s'", stage)

    # fetch the names of the files to consider separated per title
    s3_files = get_files_to_consider(config_dict)

    logger.info(
        "Collected a total of %s files, reading them...", len(s3_files.values())
    )

    logger.info("Files identified successfully, initialising the manifest.")
    # init the git repo object for the processing's repository.
    repo = git.Repo(config_dict["git_repository"])

    # ideally any not-defined param should be None.
    in_bucket = config_dict["input_bucket"]
    p_fields = config_dict["patched_fields"]
    if p_fields is not None and len(p_fields) == 0:
        p_fields = None
    prev_mft = config_dict["previous_mft_s3_path"]
    only_counting = None
    if "only_counting" in config_dict:
        only_counting = config_dict["only_counting"]

    # init the manifest given the configuration
    manifest = DataManifest(
        data_stage=stage,
        s3_output_bucket=config_dict["output_bucket"],
        s3_input_bucket=in_bucket if in_bucket != "" else None,
        git_repo=repo,
        temp_dir=config_dict["temp_directory"],
        staging=config_dict["is_staging"],
        is_patch=config_dict["is_patch"],
        patched_fields=p_fields,
        previous_mft_path=prev_mft if prev_mft != "" else None,
        only_counting=only_counting,
    )

    # processing newspapers one at a time
    for np_title, np_s3_files in s3_files.items():

        logger.info("---------- %s ----------", np_title)
        logger.debug(
            "The list of files selected for %s is: %s",
            np_title,
            np_s3_files,
        )
        # load the selected files in dask bags
        processed_files = db.read_text(
            np_s3_files, storage_options=IMPRESSO_STORAGEOPT
        ).map(json.loads)

        logger.info(
            "%s - Starting to compute the statistics on the fetched files...",
            np_title,
        )
        computed_stats = compute_stats_for_stage(processed_files, stage, client)

        logger.info(
            "%s - Populating the manifest with the resulting %s yearly statistics found...",
            np_title,
            len(computed_stats),
        )
        logger.debug("%s - computed_stats: %s", np_title, computed_stats)

        for stats in computed_stats:
            title = stats["np_id"]
            year = stats["year"]
            del stats["np_id"]
            del stats["year"]
            logger.debug("Adding %s to %s-%s", stats, title, year)
            manifest.add_by_title_year(title, year, stats)

        logger.info("%s - Finished adding stats, going to the next title...", np_title)

    logger.info("Finalizing the manifest, and computing the result...")
    # Add the note to the manifest
    if config_dict["notes"] is not None and config_dict["notes"] != "":
        manifest.append_to_notes(config_dict["notes"])
    else:
        note = f"Processing data to generate {stage} for "
        if len(config_dict["newspapers"]) != 0:
            note += f"titles: {config_dict['newspapers']}."
        else:
            note += "all newspaper titles."

        manifest.append_to_notes(note)

    if config_dict["push_to_git"]:
        manifest.compute(export_to_git_and_s3=True)
    else:
        manifest.compute(export_to_git_and_s3=False)
        manifest.validate_and_export_manifest(push_to_git=False)


def main():
    arguments = docopt(__doc__)
    config_file_path = arguments["--config-file"]
    log_file = arguments["--log-file"]
    log_level = logging.DEBUG if arguments["--verbose"] else logging.INFO
    nworkers = int(arguments["--nworkers"]) if arguments["--nworkers"] else 8
    scheduler = arguments["--scheduler"]

    init_logger(log_level, log_file)

    # suppressing botocore's verbose logging
    logging.getLogger("botocore").setLevel(logging.WARNING)
    logging.getLogger("smart_open").setLevel(logging.WARNING)

    # start the dask local cluster
    if scheduler is None:
        client = Client(n_workers=nworkers, threads_per_worker=1)
    else:
        client = Client(scheduler)

    dask_cluster_msg = f"Dask local cluster: {client}"
    logger.info(dask_cluster_msg)
    print(dask_cluster_msg)

    logger.info("Reading the arguments inside %s", config_file_path)
    with open(config_file_path, "r", encoding="utf-8") as f_in:
        config_dict = json.load(f_in)

    try:
        logger.info("Provided configuration: ")
        logger.info(config_dict)
        create_manifest(config_dict, client)

    except Exception as e:
        traceback.print_tb(e.__traceback__)
        print(e)
        client.shutdown()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding: utf-8

"""
Utility which help preparing data in view of parallel or distributed computing, in a dask-oriented view.

Usage:
    daskutils.py partition --config-file=<cf> --nb-partitions=<p> [--log-file=<f> --verbose]

Options:
    --config-file=<cf>  json configuration dict specifying various arguments
"""

import os
import logging
import docopt

from dask.diagnostics import ProgressBar
import dask.bag as db
import numpy as np

from impresso_commons.utils import init_logger
from impresso_commons.utils import Timer
from impresso_commons.path.path_s3 import s3_filter_archives
from impresso_commons.utils.s3 import get_bucket, read_jsonlines, readtext_jsonlines
from impresso_commons.utils.s3 import IMPRESSO_STORAGEOPT
from impresso_commons.utils.config_loader import PartitionerConfig

__author__ = "maudehrmann"

logger = logging.getLogger(__name__)


def partitioner(bag, path, nbpart):
    """Partition a bag into n partitions and write each partition in a file"""
    grouped_items = bag.groupby(lambda x: np.random.randint(500), npartitions=nbpart)
    items = grouped_items.map(lambda x: x[1]).flatten()
    path = os.path.join(path, "*.jsonl.bz2")
    with ProgressBar():
        items.to_textfiles(path)


def create_even_partitions(
    bucket,
    config_newspapers,
    output_dir,
    local_fs=False,
    keep_full=False,
    nb_partition=500,
):
    """Convert yearly bz2 archives to even bz2 archives, i.e. partitions.

    Enables efficient (distributed) processing, bypassing the size discrepancies of newspaper archives.
    N.B.: in resulting partitions articles are all shuffled.
    Warning: consider well the config_newspapers as it decides what will be in the partitions and loaded in memory.

    :param bucket: name of the bucket where the files to partition are
    :param config_newspapers: json dict specifying the sources to consider (name(s) of newspaper(s) and year span(s))
    :param output_dir: classic FS repository where to write the produced partitions
    :param local_fs:
    :param keep_full: whether to filter out metadata or not (i.e. keeping only text and leaving out coordinates)
    :param nb_partition: number of partitions
    :return: None
    """

    t = Timer()

    # set the output
    if local_fs:
        os.makedirs(output_dir, exist_ok=True)
        path = os.path.join(output_dir, "*.jsonl.bz2")
    else:
        path = f"{output_dir}/*.jsonl.gz"
    logger.info(f"Will write partitions to {path}")

    # collect (yearly) keys & load in bag
    bz2_keys = s3_filter_archives(bucket.name, config=config_newspapers)
    bag_bz2_keys = db.from_sequence(bz2_keys)

    # read and filter lines (1 elem = list of lines, or articles, from a key)
    if keep_full is False:
        bag_items = bag_bz2_keys.map(
            readtext_jsonlines, bucket_name=bucket.name
        ).flatten()
    else:
        bag_items = bag_bz2_keys.map(read_jsonlines, bucket_name=bucket.name).flatten()

    # repartition evenly
    grouped_items = bag_items.groupby(
        lambda x: np.random.randint(1000), npartitions=nb_partition
    )
    items = grouped_items.map(lambda x: x[1]).flatten()

    # write partitions
    with ProgressBar():
        # items.compute()
        # if local_fs:
        #     items.to_textfiles(path)
        # else:
        items.to_textfiles(path, storage_options=IMPRESSO_STORAGEOPT, compute=True)

    logger.info(f"Partitioning done in {t.stop()}.")


def main(args):
    # get args
    config_file = args["--config-file"]
    log_file = args["--log-file"]
    nb_partitions = args["--nb-partitions"]
    log_level = logging.DEBUG if args["--verbose"] else logging.INFO

    # init logger
    global logger
    logger = init_logger(logging.getLogger(), log_level, log_file)
    logger.info(f"CLI arguments received: {args}")

    config = PartitionerConfig.from_json(config_file)

    bucket = get_bucket(config.bucket_name, create=False)
    logger.info(f"Retrieved bucket: {bucket.name}")

    if args["partition"] is True:
        create_even_partitions(
            bucket,
            config.newspapers,
            config.output_dir,
            local_fs=config.local_fs,
            keep_full=config.keep_full,
            nb_partition=int(nb_partitions),
        )


if __name__ == "__main__":
    arguments = docopt.docopt(__doc__)
    main(arguments)



#!/usr/bin/python3
# coding: utf-8 

from enum import Enum
from collections import defaultdict
import os
import cv2 as cv
import numpy as np
import re
import subprocess
import sys
from subprocess import Popen, PIPE, STDOUT

__author__ = "maudehrmann"


class BoxStrategy(Enum):
    tif = "tif"
    png_highest = "png_highest"
    png_uniq = "png_uniq"
    jpg_uniq = "jpg_uniq"
    jpg_highest = "jpg_highest"


def run_cmd(cmd):
    """Execute 'cmd' in the shell and return result (stdout and stderr)."""
    try:
        p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=STDOUT, shell=True)
        result = p.communicate()

    except subprocess.CalledProcessError as e:
        sys.stderr.write(
            "common::run_command() : [ERROR]: output = %s, error code = %s\n"
            % (e.output, e.returncode))
    return result


def compose(path_img_one, path_img_two, path_img_three):
    run_cmd("composite -blend 30 " + path_img_one + " " + path_img_two + " " + path_img_three)


def get_img_from_archive(archive, path_checker, ext_checker, name_checker=None):
    if not name_checker:
        name_checker = ""
    items = sorted(
        [
            item
            for item in archive.namelist()
            if not item.startswith(".")
               and ext_checker in item
               and path_checker in item
               and name_checker in item
        ]
    )
    return items


def get_page_folders(archive):
    dirs = {os.path.dirname(item) for item in archive.namelist()}
    topdirs = {d.split("/", 1)[0] for d in dirs}
    page_folders = [os.path.join(os.path.dirname(archive.filename), page) for page in topdirs if page.isdigit()]
    return sorted(page_folders)


def get_tif(tifs, page_digit):
    # tif files have the form of 'Res/PageImg/Page0001.tif'
    for tif in tifs:
        if page_digit in tif:
            return tif


def get_jpg(jpgs, page_digit):
    # jpg files have the form of /Img/Pg006.jpg'
    # addendum: jpg files can also have various resolution: Pg010.jpg, Pg010_120.jpg, Pg010_144.jpg
    if not jpgs:
        return None
    else:
        for jpg in jpgs:
            if page_digit in jpg:
                return jpg


def get_png(pngs, page_digit):
    # png files have the form of '/Img/Pg006_180.png' or '/Img/Pg006.png'
    if not pngs:
        return None
    else:
        # group by pages (there can be several images per page)
        d = defaultdict(list)
        for i in pngs:
            elems = i.split("/", 1)
            d[elems[0]].append(elems[1])

        # take the image paths from the page we're interested in
        png_paths = d[page_digit]

        # there are several png
        if len(png_paths) > 1:
            # get pages with different resolutions (i.e. having "_": 'Img/Pg006_180.png')
            pngs_with_res = [x for x in png_paths if "_" in x]
            # build dict with k = res and v = path
            res = {}
            composite = []
            for png in pngs_with_res:
                if '_p.' in png or '_t.' in png:
                    composite.append(png)
                    continue

                r = re.search(r'_(.+)\.', png).group(1)
                res[int(r)] = png

            # sort the dict keys
            if not composite and len(res) >= 1:
                sorted_reso = sorted(res)
                # take the highest res and get the corresponding path
                return [res[sorted_reso[-1]]]

            if composite:
                sorted_composite = sorted(composite)
                return sorted_composite

        # there is only one png
        elif len(png_paths) == 1:
            return [png_paths[0]]
        else:
            return None


def get_imgdimensions(image_data):
    """ Returns image height and width"""
    img = cv.imdecode(np.frombuffer(image_data, np.uint8), 1)
    return tuple(img.shape[0:2])  # todo: check here which values to return


#!/usr/bin/env python
# coding: utf-8

import datetime
import json
import os
from enum import Enum


# TODO: implement
def index_coordinates(pages):
    """Creates an index of coordinates by text offsets.

    Gets called by `self.from_json()`.
    """
    coords = []
    for page in pages:

        for token in page['t']:
            item = {
                'page_id': page['id'],
                'coords': token['c'],
                'start_offset': token['s'],
                'end_offset': token['s'] + token['l']
            }
            if "hy1" in token:
                item['hy1'] = token['hy1']
            if "hy2" in token:
                item['hy2'] = token['hy2']
            coords.append(item)

    return coords

class ContentItemCase(Enum):
    FULL = "FULL"  # all info
    TEXT = "TEXT"  # min info + text
    LIGHT = "LIGHT"  # min info

class ContentItem:
    """
    Class which represents an impresso (rebuilt) content item.
    TODO: complement
    :ivar str id: canonical content item id
    :ivar str lg:
    :ivar str type:
    :ivar datetime date:
    :ivar str journal:
    :ivar str s3v:
    :ivar str fulltext:
    :ivar dict text_offsets: pages/regions/paragraphs/lines
    """

    def __init__(self, ci_id, lg, tp):
        """Constructor"""
        self.id = ci_id
        self.lg = lg
        self.type = tp
        self.date = self.build_date(ci_id)
        self.journal = self.build_journal(ci_id)
        self._text_offsets = {}
        self.__coordinates = []

    @staticmethod
    def build_date(ci_id):
        tmp = ci_id.split("-")
        return datetime.date(int(tmp[1]), int(tmp[2]), int(tmp[3]))

    @staticmethod
    def build_journal(ci_id):
        return ci_id.split("-")[0]

    @property
    def title(self):
        return self.__title

    @title.setter
    def title(self, value):
        self.title = value

    @property
    def lines(self):
        return self.__lines

    @lines.setter
    def lines(self, value):
        self.__lines = value

    @property
    def paragraphs(self):
        return self.__paragraphs

    @paragraphs.setter
    def paragraphs(self, value):
        self.__paragraphs = value

    @property
    def pages(self):
        return self.__pages

    @pages.setter
    def pages(self, value):
        self.__pages = value

    @property
    def regions(self):
        return self.__regions

    @regions.setter
    def regions(self, value):
        self.__regions = value

    @property
    def fulltext(self):
        return self.__fulltext

    @fulltext.setter
    def fulltext(self, value):
        self.fulltext = value

    @staticmethod
    def from_json(path=None, data=None, case=ContentItemCase.LIGHT):
        """Loads an instance of `ContentItem` from a JSON file.
        :param str path: path to a json file
        :param dict data: content item information
        :param enum case: content item configuration via `ContentItemCase`
        (LIGHT/TEXT/FULL)
        """

        assert data is not None or path is not None

        if data is not None:
            lang = None if "lg" not in data else data['lg']
            doc = ContentItem(data['id'], lang, data['tp'])
            doc.case = case

            if case == ContentItemCase.TEXT or case == ContentItemCase.FULL:
                doc.__title = data['t'] if 't' in data else None
                doc.__fulltext = data['ft'] if 'ft' in data else None

            if case == ContentItemCase.FULL:
                doc.__lines = data['lb'] if 'lb' in data else None
                doc.__paragraphs = data['pb'] if 'pb' in data else None
                doc.__regions = data['rb'] if 'pb' in data else None
                doc.__pages = data['ppreb'] if 'ppreb' in data else None
                doc.__coordinates = index_coordinates(doc.__pages)

            return doc
        elif path is not None:
            return


    def to_json(self, path, case=ContentItemCase.LIGHT):
        data = {
            "id": self.id,
            "lg": self.lg,
            "tp": self.type
        }

        if case == ContentItemCase.TEXT or case == ContentItemCase.FULL:
            data['t'] = self.__title
            data['ft'] = self.__fulltext

        if case == ContentItemCase.FULL:
            data['lb'] = self.__lines
            data['pb'] = self.__paragraphs
            data['rb'] = self.__regions
            data['ppreb'] = self.__pages

        json_output_path = os.path.join(path, f'{self.id}.json')
        try:
            with open(json_output_path, 'w') as out_file:
                json.dump(data, out_file)
            return True
        except Exception as e:
            print(e)
            return False


    def get_coordinates(self, start_offset, end_offset):
        """Enables access to image coordinates by means of text offsets.

        returns a dictionary like
        {
            'page_id': '....',
            'coordinates': [x,y,w,h]
        }
        """
        return list(filter(
            lambda c: c['start_offset'] >= start_offset and c['end_offset'] <= end_offset,
            self.__coordinates
        ))


    def __str__(self):
        s = f'{self.__class__.__name__}:\n\t' \
                f'ci_case={self.case}\n\t' \
                f'ci_id={self.id}\n\t' \
                f'ci_lg={self.lg}\n\t' \
                f'ci_type={self.type}\n\t' \
                f'ci_date={self.date}\n\t' \

        if self.case == ContentItemCase.TEXT \
                or self.case == ContentItemCase.FULL:
            s = s + f'ci_fulltext={self.fulltext}\n\t' \
                   f'ci_title={self.title}\n\t' \

        return s
